# 
# math.py                                                               
# 
# D. Clarke
# 
# Wrappers for some math functions, some math functions that I couldn't find in numpy or scipy, and some methods
# for comparing math objects. Some functions already existing in numpy or scipy may be implemented when I want
# to expose what those functions do under the hood.
#


import numpy as np
import scipy as sp
import latqcdtools.base.logger as logger
from latqcdtools.base.utilities import isArrayLike, cleanOutput
from latqcdtools.base.check import checkType


# ----------------------------------------------------------------- MATRICES


def id(N) -> np.ndarray:
    """ 
    NxN complex identity matrix
    """
    checkType('int',N=N)
    return np.eye(N,dtype=complex)


def ze(N) -> np.ndarray:
    """ 
    NxN complex zero matrix
    """
    checkType('int',N=N)
    return np.zeros((N,N), dtype=complex)


def isMatrix(mat) -> bool:
    checkType(np.ndarray,mat=mat)
    return mat.ndim==2


def checkMatrix(mat):
    if not isMatrix(mat):
        logger.TBRaise('Expected matrix. Got shape',np.shape(mat))


def isVector(vec) -> bool:
    checkType(np.ndarray,mat=vec)
    return vec.ndim==1


def checkVector(vec):
    if not isVector(vec):
        logger.TBRaise('Expected vector. Got shape',np.shape(vec))


def isSquare(mat) -> bool:
    checkMatrix(mat)
    return mat.shape[0] == mat.shape[1]


def checkSquare(mat):
    """ 
    Make sure mat is a square np.ndarray object. 

    Args:
        mat (np.ndarray)
    """
    if not isSquare(mat):
        logger.TBRaise('Expected square matrix. Got shape',np.shape(mat))


def dagger(arr) -> np.ndarray:
    checkType(np.ndarray,mat=arr)
    if arr.ndim > 2:
        logger.TBRaise('Expected ndim < 3. Got ndim =',arr.ndim) 
    return np.conjugate(arr).T


def isUnitary(mat) -> bool:
    """
    Unitary matrices U satisfy U^dag U = 1.
    
    Args:
        mat (np.ndarray)

    Returns:
        bool: True if unitary
    """
    checkMatrix(mat)
    N = len(mat[0])
    return rel_check(id(N),mat @ dagger(mat))


def isSpecial(mat) -> bool:
    """
    Special matrices M satisfy det(M) = 1.

    Args:
        mat (np.ndarray)

    Returns:
        bool: True if special
    """
    checkSquare(mat)
    return rel_check(np.linalg.det(mat), 1.)


def isSymmetric(mat) -> bool:
    """
    Symmetric matrices satisfy M^t = M.

    Args:
        mat (np.ndarray)

    Returns:
        bool: True if symmetric
    """
    checkSquare(mat)
    return np.allclose(mat, mat.T)


def isHermitian(mat) -> bool:
    """
    Hermitian matrices satisfy dagger(M)=M

    Args:
        mat (np.ndarray)

    Returns:
        bool: True if hermitian 
    """
    checkSquare(mat)
    return np.allclose(mat, dagger(mat))


def isAntihermitian(mat) -> bool:
    """
    Antihermitian matrices satisfy dagger(M)=-M

    Args:
        mat (np.ndarray)

    Returns:
        bool: True if antihermitian 
    """
    checkSquare(mat)
    return np.allclose(mat, -dagger(mat))


def isOrthogonal(mat) -> bool:
    """
    Orthogonal matrices satisfy M^t M = id

    Args:
        mat (np.ndarray)

    Returns:
        bool: True if orthogonal 
    """
    checkSquare(mat)
    N = len(mat[0])
    return np.allclose(mat@mat.T, id(N))


def isHankel(mat) -> bool:
    """
    Hankel matrices look like (3d example)
    a b c
    b c d
    c d e

    Args:
        mat np.ndarray 

    Returns:
        bool: True if Hankel 
    """
    checkSquare(mat)
    hankel = True
    for j in range(mat.shape[0]):
        for i in range(j):
            for k in range(j-i):
                hankel *= mat[i,j]==mat[i+k,j-k]
    return hankel


def TA(mat) -> np.ndarray:
    """
    Make mat traceless and antihermitian

    Args:
        mat (np.ndarray)

    Returns:
        bool: _description_
    """
    checkSquare(mat)
    return 0.5*(mat-dagger(mat))


def isPositiveSemidefinite(mat,details=False,eps=1e-12) -> bool:
    """ Returns true if mat is positive semidefinite. Otherwise, if details=True,
    list the eigenvalues that are not >=0.

    Args:
        mat (np.ndarray)
        details (bool, optional): If true, report problematic eigenvalues. Defaults to False.

    Returns:
        bool: True if positive semidefinite 
    """
    checkSquare(mat)
    eigenvalues = np.linalg.eigvals(mat)
    positiveSemidefinite = np.all(eigenvalues >= -eps)
    if details and (not positiveSemidefinite):
        logger.info('Problem eigenvalues:')
        for i in range(len(eigenvalues)):
            if not eigenvalues[i]>=0:
                logger.info(f'  lambda[{i}] = {eigenvalues[i]}')
    return positiveSemidefinite


def regulate(mat,svdcut=1e-12) -> np.ndarray:
    """ 
    If a matrix's singular values are too small, it will be ill-conditioned,
    making it difficult to invert and hence reducing numerical stability. This method
    extracts its singular values using SVD, then doctors the singular values to reduce
    the condition number. In the context of applying an SVD cut to a covariance 
    matrix, see e.g. Appendix D of 10.1103/PhysRevD.100.094508.

    Args:
        mat (np.ndarray)
        svdcut (float, optional): condition number threshold. Defaults to 1e-12.

    Returns:
        np.ndarray: regulated matrix 
    """
    logger.debug('regulate matrix with condition number',cleanOutput(np.linalg.cond(mat)))
    U, s, Vdagger = sp.linalg.svd(mat)
    smax = np.max(np.linalg.eigvals(mat))
    s[s < svdcut*smax] = svdcut*smax
    res = U @ np.diag(s) @ Vdagger 
    logger.debug('new condition number:',cleanOutput(np.linalg.cond(res)))
    return res


def invert(mat,method='scipy',svdcut=1e-12) -> np.ndarray:
    """ 
    Invert matrix.

    Args:
        mat (np.ndarray): to-be-inverted matrix 
        method (str): algorithm for inverting the matrix
        
    Returns:
        np.ndarray: mat^{-1} 
    """
    checkType(np.ndarray,mat=mat)
    checkSquare(mat)
    checkType(str,method=method)
    if method=='scipy':
        return sp.linalg.inv(mat)
    elif method=='numpy': # Seems to be quite slow
        return np.linalg.inv(mat)
    elif method=='pinv': 
        return np.linalg.pinv(mat)
    elif method=='svd':
        U, s, Vdagger = sp.linalg.svd(mat)
        return dagger(Vdagger) @ np.diag(1/s) @ dagger(U)
    elif method=='auto':
        condition_number = np.linalg.cond(mat)
        if condition_number > 1/svdcut:
            res = invert(regulate(mat,svdcut=svdcut),'pinv')
            return res
        else:
            return invert(mat,'scipy')
    else:
        logger.TBRaise('Unrecognized inverter',method)


def logDet(mat) -> float:
    """ 
    Logarithm of determinant. 
    """
    checkType(np.ndarray,mat=mat)
    checkSquare(mat)
    _, ans = np.linalg.slogdet(mat)
    return ans


def exp(mat) -> np.ndarray:
    """
    Exponential of square matrix.
    """
    checkSquare(mat) 
    return sp.linalg.expm(mat)


def log(mat) -> np.ndarray:
    """
    Natural logarithm of square matrix.
    """
    checkSquare(mat)
    return sp.linalg.logm(mat)


def pow(mat,power) -> np.ndarray:
    """
    Matrix power. 
    """
    checkSquare(mat)
    checkType('int',power=power)
    return np.linalg.matrix_power(mat, power)


class SUN(np.ndarray):

    """ 
    A member of the Lie group SU(N). Implemented as a subclass of the np.ndarray class. This gives us access already
    to all the nice features of np.ndarray and lets us leverage the speed of numpy.
        g.trace()
        g.det()
        g.dagger()
        g[i,j], which can be used to access and assign
        g + h
        g*h = g@h
        2*g
    """

    def __new__(cls, N=None, mat=None):
        if N is None:
            logger.TBRaise('Must specify N')
        if mat is None:
            mat = ze(N) 
        obj = np.asarray(mat, dtype=complex)
        if obj.shape != (N, N):
            logger.TBRaise(f"SU({N}) matrices must have shape ({N},{N})")
        return np.copy(obj).view(cls)


    def __repr__(self) -> str:
        return "SU(N)"


    def __mul__(self, other):
        # Perform matrix multiplication instead of element-wise multiplication
        return np.dot(self, other)


    def __pow__(self,power):
        # Perform matrix power instead of element-wise power
        return pow(self, power)


    def trace(self):
        return super().trace().item()


    def dagger(self):
        return dagger(self) 


    def det(self):
        return np.linalg.det(self)


    def isSUN(self) -> bool:
        """ 
        Check that I have det=1 and am unitary. 
        """
        if not (isSpecial(self) and isUnitary(self)):
            return False
        return True


    def setToMatrix(self,other):
        """ 
        Turn into RHS link. 
        """
        np.copyto(self,np.asarray(other,dtype=complex))


    def setToZero(self):
        """ 
        Turn into zero matrix. 
        """
        N = len(self)
        self.setToMatrix(ze(N))


    def setToIdentity(self):
        """ 
        Turn into identity matrix. 
        """
        N = len(self)
        self.setToMatrix(id(N)) 


# --------------------------------------------------------------- OTHER MATH


def pnorm(arr,p=2) -> float:
    """
    Returns p-norm of vector or matrix arr.

    Args:
        arr (np.ndarray)
        p (float, optional): Defaults to 2.

    Returns:
        float: pnorm 
    """
    checkType(np.ndarray,arr=arr)
    checkType("real",p=p)
    if p < 1:
        logger.TBRaise('p-norm defined only for p>=1.')
    if arr.ndim > 2: 
        logger.TBRaise('Expected ndim < 3. Got ndim =',arr.ndim) 
    if p==1:
        if isMatrix(arr):
            res = np.max(np.sum(np.abs(arr), axis=0))
        elif isVector(arr):
            res = np.sum(np.abs(arr))
    elif p==2:
        if isMatrix(arr):
            res = np.sqrt(np.trace(dagger(arr) @ arr)).real
        elif isVector(arr):
            res = np.sqrt(         dagger(arr) @ arr).real
    elif p==np.inf:
        if isMatrix(arr):
            res = np.max(np.sum(np.abs(arr), axis=1))
        elif isVector(arr):
            res = np.max(np.abs(arr))
    else:
        res = np.sum(np.abs(arr)**p)**(1./p) 
    return res


def normalize(arr,p=2) -> np.ndarray:
    """
    Normalize vector or matrix arr using p-norm.

    Args:
        arr (np.ndarray)
        p (float, optional): Defaults to 2.

    Returns:
        np.ndarray: normalized array 
    """
    checkType(np.ndarray,arr=arr)
    return arr/pnorm(arr,p)


def fallFactorial(n,m) -> float:
    """ 
    Falling factorial n fall to m. 
    """
    checkType('int',n=n)
    checkType('int',m=m)
    if m>n:
        logger.TBRaise("m>n.")
    return sp.special.poch(n-m+1,m)


def riseFactorial(n,m) -> float:
    """ 
    Rising factorial n rise to m. 
    """
    checkType('int',n=n)
    checkType('int',m=m)
    if n>m:
        logger.TBRaise("n>m.")
    return sp.special.poch(n,m)


def RMS(data) -> float:
    """
    Root-mean-square of data

    Args:
        data (np.ndarray)

    Returns:
        float: RMS of data 
    """
    checkType(np.ndarray,data=data)
    return np.sqrt(np.mean(data**2))


def quadrature(data) -> float:
    """
    Add data in quadrature

    Args:
        data (np.ndarray)

    Returns:
        float: data added in quadrature 
    """
    checkType(np.ndarray,data=data)
    return np.sqrt(np.sum(data**2))


def rel_check(a, b, prec = 1e-6, abs_prec = 1e-14) -> bool:
    """ 
    Check whether a and b are equal. a and b can be array-like, float-like, or complexes. If a
    and b are array-like, we check that they are element-wise equal within the tolerance. 

    Args:
        a (obj)
        b (obj)
        prec (float, optional): Relative precision. Defaults to 1e-6.
        abs_prec (float, optional): Absolute precision. Defaults to 1e-14.

    Returns:
        bool: True if a and b are equal. 
    """
    if isArrayLike(a):
        if np.shape(a) != np.shape(b):
            logger.TBRaise('a and b must have the same shape. Received a, b shapes =',np.shape(a),np.shape(b))
        return np.allclose( a, b, rtol = prec, atol = abs_prec)
    else:
        checkType("scalar",a=a)
        checkType("scalar",b=b)
        try:
            return np.isclose( a, b, rtol = prec, atol = abs_prec)
        except TypeError:
            logger.TBRaise('Expected reals, complexes, or array-like. Received a, b types =',type(a),',',type(b),exception=TypeError)



#
# jackknife.py
#
# H. Dick, D. Clarke
#


import numpy as np
from latqcdtools.statistics.statistics import std_mean, std_err
import latqcdtools.base.logger as logger
from latqcdtools.base.speedify import parallel_function_eval
from latqcdtools.base.check import checkType
from latqcdtools.base.utilities import isHigherDimensional


def _pseudobins(jackknifeBins,avg):
    """ 
    Calculate the 'pseudovalue' from the ith jackknife estimator. The pseudovalue is unbiased
    up to O(1/N), where N is the number of data. See e.g. eq. (1.1) of Miller, Biometrika 1974. 
    """ 
    nblocks=len(jackknifeBins)
    return nblocks*np.array(avg) - (nblocks-1)*jackknifeBins


def _pareAxis(data,axis,nblocks):
    """ 
    In order to ensure that all jackknife blocks have the same length, we pare the data along
    the conf_axis, so that nblocks divides the length of data along conf_axis.
    """
    data=np.array(data)
    length = data.shape[axis]
    new_length = length - (length % nblocks)
    indices = [slice(None)] * data.ndim
    indices[axis] = slice(new_length)
    return data[tuple(indices)]


def jackknife(f, data, numb_blocks=20, conf_axis=1, nproc=1, return_sample=False, args=()):
    """ 
    Carry out a jackknife of an arbitrary function f of some data.

    Args:
        f (func)
        data (array-like)
        numb_blocks (int, optional): Number of jackknife blocks. Defaults to 20.
        conf_axis (int, optional): The axis that represents the configuration axis, i.e., measurements
          are along this axis. Defaults to 1 when data has dimension of at least 2; else defaults to 0.
        nproc (int, optional): Number of threads to use. Defaults to 1. Increasing nproc is likely to
          have a deleterious impact on performance when f is a fast function. When f is slower, for
          example if you are doing curve fits or integrating in the block, increasing nproc may
          improve performance. Feel free to benchmark it for your use-case. 
        return_sample (bool, optional): Return pseudovalues? Defaults to False.
        args (tuple, optional)

    Returns:
        jackknife mean and error (optionally pseudovalues) 
    """
    checkType('int',numb_blocks=numb_blocks)
    checkType('int',conf_axis=conf_axis)
    checkType('int',nproc=nproc)
    checkType(bool,return_sample=return_sample)
    if type(data) is tuple:
        logger.TBRaise('Does not support tuple data. Use latqcdtools.legacy.jackknife if you need this.')
    if numb_blocks <= 1:
        logger.TBRaise('Need numb_blocks > 1. Set numb_blocks=len(data) for remove-1 jackknife.')
    if not isHigherDimensional(data):
        conf_axis=0
    data = np.array(data)
    if numb_blocks > np.size(data,axis=conf_axis):
        logger.TBRaise(f'Need numb_blocks <= ndata. numb_blocks, ndata = {numb_blocks}, {np.size(data,axis=conf_axis)}')
    data = _pareAxis(data,conf_axis,numb_blocks)
    n = data.shape[conf_axis]
    total = f(data, *args)
    block_id = np.linspace(0, numb_blocks, n, endpoint=False).astype(np.int32)
    def fJ(i):
        sub_data = np.compress((block_id != i), data, axis=conf_axis)
        return f(sub_data, *args)
    J = np.array(parallel_function_eval(fJ,range(numb_blocks),nproc=nproc))
    bins = _pseudobins(J,total)
    if return_sample:
        return bins, std_mean(bins), std_err(bins) 
    else:
        return std_mean(bins), std_err(bins) 



#
# utilities.py
#
# D. Clarke
#
# Some utilities that you might use in any program.
#

from subprocess import run, PIPE
import numpy as np
import time, re, datetime, os, shutil
import latqcdtools.base.logger as logger
import glob


# For byte conversions
_bytePrefix = { "Q"  : 1024**10,
                "R"  : 1024**9,
                "Y"  : 1024**8,  
                "Z"  : 1024**7, 
                "E"  : 1024**6, 
                "P"  : 1024**5, 
                "T"  : 1024**4, 
                "G"  : 1024**3, 
                "M"  : 1024**2, 
                "k"  : 1024, 
                "1"  : 1,
                1    : 1 }


def _getPrefix(byteString):
    if byteString=="B": 
        prefix=1
    else:
        prefix=byteString[0]
    return prefix


def _convert(text):
    if text.isdigit():
        return int(text)
    else:
        return text.lower()


def _alphanum_key(key):
    """ 
    Splits the string `key` at any point where there is one or more consecutive digits. 
    The regular expression `([0-9]+)` is used to match one or more digits. The parentheses `()` 
    capture the matched digits as separate elements. For example, if `key` were 
    `'abc123def456ghi'`, the resulting list would be `['abc', '123', 'def', '456', 'ghi']`. 
    """
    return [_convert(c) for c in re.split('([0-9]+)', key)]


# ---------------------------------------------------------------------------------- MAKE INTERNAL FUNCTIONS MORE SMOOTH


def isArrayLike(obj) -> bool:
    """ 
    Figure out whether obj is indexable.

    Args:
        obj (python object)

    Returns:
        bool: True if there is at least one index, false otherwise. 
    """
    try:
        obj[0]
        return True
    except (TypeError,IndexError):
        return False


def isHigherDimensional(obj) -> bool:
    """ 
    Figure out whether obj has at least two indices.

    Args:
        obj (array-like)

    Returns:
        bool: True if there are at least two indices, false otherwise. 
    """
    try:
        obj[0][0]
        return True
    except (TypeError, IndexError):
        return False


def isIntType(obj) -> bool:
    return isinstance(obj,(int, np.int8, np.int16, np.int32, np.int64))


def isFloatType(obj) -> bool:
    return isinstance(obj,(float,np.float16,np.float32,np.float64,np.float128))


def isComplexType(obj) -> bool:
    return isinstance(obj,(complex,np.complex64,np.complex128))


def isScalar(obj) -> bool:
    if (not isIntType(obj)) and (not isFloatType(obj)) and (not isComplexType(obj)):
        return False
    return True


def unvector(obj):
    """ 
    Remove outermost brackets of array-like object with single element, if possible. This is needed
    because sometimes different numpy methods give inconsistent outputs, like turning a scalar
    into a zero-dimensional array, a 1-dimensional array, or just the scalar itself.

    Args:
        obj (python object)

    Returns:
        obj, obj[0], or obj.item() depending on obj
    
    """ 
    if isinstance(obj,np.ndarray):
        if obj.ndim==0:
            return obj.item()
    if not isArrayLike(obj):
        return obj
    N = len(obj)
    if N > 1:
        return obj
    else:
        return obj[0]


def envector(*args):
    """ 
    Change obj to a numpy array if it's a scalar. Sometimes required when, e.g., using np.vectorize. 
    """
    result = ()
    for obj in args:
        if not isArrayLike(obj):
            obj = np.array([obj])
        result += (obj,)
    return unvector(result)


def toNumpy(*args,**kwargs):
    result = ()
    for obj in args:
        if isArrayLike(obj):
            obj = np.array(obj,**kwargs)
        result += (obj,)
    return result


def appendToDocstring(string=None,args=None,returns=None):
    def decorator(func):
        appendand=''
        if string is not None:
            appendand += string
        if args is not None:
            appendand += """\n    Args:""" + args
        if returns is not None:
            appendand += """\n    Returns:""" + returns
        if func.__doc__ is None:
            func.__doc__ = appendand 
        else:
            func.__doc__ = func.__doc__ + appendand 
        return func
    return decorator


# ------------------------------------------------------------------------------------------------- CONVENIENCE FOR USER


def getArgs(parser):
    """ 
    Get arguments from the ArgumentParser. Complain if you don't get exactly the correct arguments. 
    """
    args, invalid_args = parser.parse_known_args()
    if len(invalid_args)>0:
        logger.TBRaise("Received unrecognized arguments",invalid_args,".")
    return args


def printArg(message,param):
    """ 
    Some arguments are None by default, and you only want to print them if they are set. 
    """
    if param is not None:
        logger.info(message,param)


def cleanOutput(*args,label=None) -> str:
    """ 
    This method takes a bunch of args and formats them automatically for output. The idea is
    that you can use this method to ensure that columns are well lined up.

    Args:
        *args: The numbers you want to output, separated by commas. 
        label (str, optional): Put label to the left of your output. Defaults to None.

    Returns:
        str: formatted output string 
    """
    data = ()
    form = ''
    if label is not None:
        if not isinstance(label,str):
            logger.TBRaise('label must be a string')
        form += '%'+str(len(label))+'s'
        data += (label,)
    spacing = ''
    for col in args:
        if col is None:
            data += ('',)
            form += spacing+'%15s'
        elif isinstance(col,str):
            data += (col,)
            form += spacing+'%20s'
        elif isinstance(col,complex):
            data += (col.real,)
            data += (col.imag,)
            form += spacing+'%15.8e  %15.8e'
        elif isinstance(col,list):
            logger.TBRaise('Expected list of scalars rather than list of lists.')
        else:
            data += (col,)
            form += spacing+'%15.8e'
        spacing = '  '
    return form % data


def printClean(*args,label=None):
    """ 
    Wrapper for cleanOutput that prints to screen.

    Args:
        *args: The numbers you want to output, separated by commas. 
        label (str, optional): Put label to the left of your output. Defaults to None.
    """
    logger.info(cleanOutput(*args,label).strip())


def printDict(dic,level=0):
    """ 
    Prints key, value pairs line by line. 
    """
    if not isinstance(dic,dict):
        logger.TBRaise('Expected type', dict, 'but received', type(dic))
    indent='  '*level
    for key in dic:
        if type(dic[key])==dict:
            logger.info(f'{indent}{key}:')
            printDict(dic[key],level+1)
        else:
            printClean(f'{indent}{key}:',dic[key])


def shell(*args):
    """ 
    Carry out the passed arguments args in the shell. Can be passed as a single
    string or as a list. Captures and returns output of shell command. E.g.
        shell('ls -lah')
    """
    args = [str(s) for s in args]
    process = run(' '.join(args),shell=True,check=True,stdout=PIPE,universal_newlines=True)
    return process.stdout


def shellVerbose(*args):
    """ 
    Same as shell, but instead of capturing output, print it to screen. 
    """
    args = [str(s) for s in args]
    process = run(' '.join(args),shell=True,check=True,stdout=PIPE,universal_newlines=True)
    print(process.stdout)


def comesBefore(date1,date2,format="%Y/%m/%d %H:%M:%S",beforeOrEqual=False) -> bool:
    """ 
    Check whether date1 comes before date2.

    Args:
        date1 (str)
        date2 (str)
        format (str): format for date strings. Defaults to "%Y/%m/%d %H:%M:%S"
        beforeOrEqual (bool): also return True if date1 == date2
        
    Returns:
        bool: date1 < date2 
    """
    date1_converted = datetime.datetime.strptime(date1, format)
    date2_converted = datetime.datetime.strptime(date2, format)
    if beforeOrEqual:
        return date1_converted <= date2_converted
    return date1_converted < date2_converted


def elapsedSeconds(date1,date2,format="%Y/%m/%d %H:%M:%S") -> float:
    """
    Compute elapsed time in seconds between date1 and date2. 

    Args:
        date1 (str)
        date2 (str)
        format (str): format for date strings. Defaults to "%Y/%m/%d %H:%M:%S"
        
    Returns:
        float: elapsed time in seconds 
    """
    date1_converted = datetime.datetime.strptime(date1, format)
    date2_converted = datetime.datetime.strptime(date2, format)
    elapsed = date1_converted-date2_converted
    return abs(elapsed.total_seconds())


def naturalSort(l) -> list:
    """
    Sort list of strings so that, e.g. '10' comes after '9' rather than before it.
    """
    return sorted(l, key=_alphanum_key)


def find_nearest_idx(array, value) -> int:
    """ 
    Find the index of the element of array nearest to value. 
    """
    array = np.array(array)
    idx = (np.abs(array - value)).argmin()
    return idx


def substringBetween(string,a,b) -> str:
    """ 
    Find the substring of string between a and b. If a==b, it looks between the
    first and second occurences of a. 

    Args:
        string (str)
        a (str): starting delimiter 
        b (str): ending delimiter

    Returns:
        str: substring
    """
    start_index = string.index(a) + len(a)
    end_index   = string[start_index:].index(b) + start_index
    return string[start_index:end_index]


def deleteLine(target,line_number):
    """
    Delete line line_number from file target.

    Args:
        target (str)
        line_number (int)
    """
    if os.path.isfile(target):
        with open(target, 'r') as file:
            lines = file.readlines()
        if 0 < line_number <= len(lines):
            lines.pop(line_number - 1)  # line_number is 1-based
            with open(target, 'w') as file:
                file.writelines(lines)
        else:
            logger.TBRaise("Line number is out of range.")
        return
    logger.warn(f"{target} does not exist.")


def deleteFile(target):
    """ 
    Delete the file at target, if it exists. 
    """
    if os.path.isfile(target):
        try:
            os.remove(target)
            logger.info("Deleted file",target)
            return
        except OSError:
            pass
    else:
        pass
    logger.warn('Unable to remove file',target)


def deleteFolder(target):
    """ 
    Delete the folder at target, if it exists. 
    """
    if os.path.isdir(target):
        try:
            shutil.rmtree(target)
            logger.info("Deleted folder",target,"and its subdirectories.")
            return
        except OSError:
            pass
    else:
        pass
    logger.warn('Unable to remove folder',target)


def createFilePath(fullFileName):
    """ 
    Create the directory path if it isn't there already. 
    """
    if '/' in fullFileName:
        dir_path = os.path.dirname(fullFileName)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path,exist_ok=True)


def ls(filePath) -> list:
    return naturalSort(list(glob.iglob(filePath)))


def byteConvert(x,b1,b2):
    """ 
    Convert between bytes given scientific prefixes.

    Args:
        x (float): Bytes in original units. 
        b1 (str): Original units. 
        b2 (str): Target units.

    Returns:
        float: Bytes in target units. 
    """
    p1 =_getPrefix(b1)
    p2 =_getPrefix(b2)
    num=_bytePrefix[p1]
    den=_bytePrefix[p2]
    return x*num/den


class timer:

    """ 
    A class to facilitate doing rudimentary timings in the Toolbox. 
    """

    def __init__(self):
        logger.info("Timer initialized.")
        self._tstart = time.time()
        self._tend   = self._tstart

    def __repr__(self) -> str:
        return "timer"

    def printTiming(self, message=None):
        self._tstart = self._tend
        self._tend   = time.time()
        timing = self._tend - self._tstart
        if message is None:
            logger.info("Time to finish: %12.8f [s]." % timing)
        else:
            logger.info("Time to finish "+message+": %12.8f [s]." % timing)

    def resetTimer(self):
        self._tstart = time.time()
        self._tend   = self._tstart


# 
# readWrite.py
#
# D. Clarke
#
# Methods for convenient reading and writing. 
#


import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkType
from latqcdtools.base.cleanData import clipRange, excludeAtCol 
from latqcdtools.base.utilities import createFilePath, isComplexType


def readTable(filename,unpack=True,col=None,minVal=-np.inf,maxVal=np.inf,excludeAtVal=None,**kwargs) -> np.ndarray:
    """ 
    Wrapper for np.loadtxt. It unpacks by default to prevent transposition errors, and also optionally
    allows the user to restrict the table based on the range of one of the columns.

    Args:
        filename (str):
            Name of the file to open. 
        unpack (bool, optional): 
            If False, reads the table in a confusing, useless way. Defaults to True.
        col (int, optional): 
            Use this column to restrict the range of the rest of the table. Defaults to None.
        minVal (float, optional): 
            Minimum value for above restriction. Defaults to -np.inf.
        maxVal (float, optional):
            Maximum value for above restriction. Defaults to np.inf.
        excludeAtVal (float, optional):
            Throw out all rows for which column col has value excludeAtVal. Exclusion is
            carried out after range restriction. 
            
    Returns:
        np.array: Data table. 
    """
    checkType(str,filename=filename)
    try: 
        data = np.loadtxt(filename,unpack=unpack,**kwargs)
    except Exception as e:
        raise e
    if col is not None:
        data = clipRange(data,col=col,minVal=minVal,maxVal=maxVal)
        if excludeAtVal is not None:
            data = excludeAtCol(data,col=col,atVal=excludeAtVal)
    else:
        if minVal!=-np.inf:
            logger.TBRaise('Set col=None with minVal')
        if maxVal!=np.inf:
            logger.TBRaise('Set col=None with maxVal')
        if excludeAtVal is not None:
            logger.TBRaise('Set col=None with excludeAtVal')
    return data


def writeTable(filename,*args,**kwargs):
    """ 
    Wrapper for np.savetxt. The idea is that you can use it like this:
    
    writeTable('file.txt',col1,col2,header=['header1','header2])
    
    This works for an arbitrary number of 1-d columns col. It seems much more intuitive to me 
    that you pass columns as arguments than whatever np.savetxt is doing.

    Args:
        filename (str): output file name
    """
    npkwargs=kwargs
    if len(args)==0:
        logger.TBRaise('No data passed to writeTable.')
    width=15
    prec=8
    if 'width' in kwargs:
        width = kwargs['width']
        del npkwargs['width']
    if 'digits' in kwargs:
        prec = kwargs['digits']
        del npkwargs['digits']
    if 'header' in kwargs:
        head = kwargs['header']
        del npkwargs['header']
        checkType(list,header=head)
        if len(head)!=len(args):
            logger.TBRaise(f'Mismatch between header length ({len(head)}) and number of columns ({len(args)})')
    else:
        head = None

    data = ()
    dtypes = []
    form = ''
    hform = ''
    hstr = tuple()
    colno = 0
    ndat = len(args[0])

    for i,col in enumerate(args):

        col_arr = np.array(col)
        if len(col_arr) != ndat:
            logger.TBRaise('Expected length',ndat,'for col',colno,'but found',len(col_arr))

        if isComplexType(col_arr[0]): 
            data += (col_arr.real,)
            data += (col_arr.imag,)
            form += f'  %{width}.{prec}e  %{width}.{prec}e'
            dtypes.append( (_lab(colno), float) )
            dtypes.append( (_lab(colno+1), float) )
            colno += 2
            if head is not None:
                hstr += ('Re '+head[i],'Im '+head[i],)
                if hform=='':
                    hform += f'%{width}s  %{width}s'
                else:
                    hform += f'  %{width}s  %{width}s'
        elif isinstance(col_arr[0],str):
            data += (col_arr,)
            form += f'  %{width}s'
            dtypes.append( (_lab(colno), f'U{width}' ) )
            colno += 1
            if head is not None:
                hstr += (head[i],)
                if hform=='':
                    hform += f'%{width}s'
                else:
                    hform += f'  %{width}s'
        else:
            data += (col_arr,)
            form += f'  %{width}.{prec}e'
            dtypes.append( (_lab(colno), float) )
            colno += 1
            if head is not None:
                hstr += (head[i],)
                if hform=='':
                    hform += f'%{width}s'
                else:
                    hform += f'  %{width}s'

    if head is not None:
        head = hform % hstr 
    ab = np.zeros(data[0].size, dtype=dtypes)
    for i in range(colno):
        ab[_lab(i)] = data[i]
    createFilePath(filename)
    if head is None:
        np.savetxt(filename, ab, fmt=form, **kwargs)
    else:
        np.savetxt(filename, ab, fmt=form, header=head, **kwargs)


def _lab(num) -> str:
    """ 
    Create a short string label for each column of a data table. Needed for writeTable. 
    """
    return 'var' + str(num)



# 
# constants.py
# 
# D. Clarke 
# 
# Many constants and unit conversions that are relevant for physics. These values do not necessarily
# represent the most up-to-date or appropriate choices; rather they are a collection of constants I
# have needed for work or teaching classes. This is especially true of constants like particle masses,
# where choosing one value over another may be subtle. 
#

import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkType
from latqcdtools.math.math import quadrature


# Base constants for unit conversions
hceVm           = 197.3269788e-9   # hbar c in [eV m]. PDG 2018. DOI: 10.1103/PhysRevD.98.030001.
cms             = 299792458        # c in [m/s]. NIST 2018.
days_per_year   = 365.2422         # NIST 2018.
days_per_month  = days_per_year/12
meters_per_mile = 1609.344         # NIST 2023 based on international foot (not survey foot).
BTU_per_Wh      = 3.412            # Many possible definitions--don't take too seriously. A stupid unit, indeed.
kBJdivK         = 1.380649e-23     # kB in [J/K]. NIST 2018.
eC              = 1.602176634e-19  # e in [C]. NIST 2018.


# See corresponding _phys() functions for the references
_fkerrs2012   = np.array([0.2   ,0.8   ,0.2   ])
_fkerrs2018   = np.array([0.17  ,0.45  ,0.16  ])
_fpierrs2018  = np.array([0.01  ,0.03  ,0.13  ])
_r1fmerrs2010 = np.array([0.0008,0.0014,0.0004])


# List of scientific prefixes
_prefix = { "Q"  : 1e30,
            "R"  : 1e27,
            "Y"  : 1e24,  
            "Z"  : 1e21, 
            "E"  : 1e18, 
            "P"  : 1e15, 
            "T"  : 1e12, 
            "G"  : 1e9, 
            "M"  : 1e6, 
            "k"  : 1e3, 
            "h"  : 1e2, 
            "da" : 1e1, 
            "1"  : 1,
            1    : 1,
            "d"  : 1e-1,
            "c"  : 1e-2, 
            "m"  : 1e-3, 
            "u"  : 1e-6, 
            "n"  : 1e-9, 
            "p"  : 1e-12, 
            "f"  : 1e-15, 
            "a"  : 1e-18, 
            "z"  : 1e-21, 
            "y"  : 1e-24, 
            "r"  : 1e-27, 
            "q"  : 1e-30  }


# Alphabetical order for easier finding
_baseUnits = [
             "BTU",    # British thermal unit
             "degC",   # Celsius
             "degF",   # Fahrenheit
             "eV",     # electron-volt
             "ft",     # feet
             "h",      # hour
             "J",      # Joule
             "K",      # Kelvin
             "m",      # meter
             "m/s",    # meter/second
             "mi",     # mile
             "mi/h",   # mile/hour
             "min",    # minute
             "s",      # second
             "d",      # day
             "W",      # Watt
             "Wh",     # Watt-hour
             "Wh/y",   # Watt-hour/year
             "y",      # year
             ]


def _separatePrefix(units):
    checkType(str,units=units)
    if units in _baseUnits: 
        prefix=1
        baseUnit=units
    else:
        prefix=units[0]
        baseUnit=units[1:]
    return prefix, baseUnit


def convert(x,unit1,unit2) -> float:
    """ 
    General method for doing unit conversions. He knows about scientific prefixes like G, M, and so on.
    If the unit ends in 'inv', it is interpreted as 1/unit.

    Args:
        x (float): measurement in [unit1]. 
        unit1 (str): Original units.
        unit2 (str): Target units.

    Returns:
        float: measurement in [unit2]. 
    """
    p1, u1 = _separatePrefix(unit1)
    p2, u2 = _separatePrefix(unit2)
    if not p1 in _prefix:
        logger.TBRaise('Unknown prefix',p1)
    if not p2 in _prefix:
        logger.TBRaise('Unknown prefix',p2)
    u1u2 = (u1,u2)

    if u1.endswith('inv'):
        num=1/_prefix[p1]
    else:
        num=_prefix[p1]
    if u2.endswith('inv'):
        den=1/_prefix[p2]
    else:
        den=_prefix[p2]
    fac = num/den

    if u1==u2:
        result = x

    # time 
    elif u1u2==('s','min'):
        result = x/60
    elif u1u2==('min','s'):
        result = x*60
    elif u1u2==('s','h'):
        result = x/60**2
    elif u1u2==('h','s'):
        result = x*60**2
    elif u1u2==('s','y'):
        result = x/(days_per_year*24*60**2)
    elif u1u2==('y','s'):
        result = x*(days_per_year*24*60**2)
    elif u1u2==('h','y'):
        result = x/(days_per_year*24)
    elif u1u2==('y','h'):
        result = x*(days_per_year*24)
    elif u1u2==('d','y'):
        result = x/days_per_year
    elif u1u2==('y','d'):
        result = x*days_per_year

    # distance
    elif u1u2==('mi','m'):
        result = x*meters_per_mile
    elif u1u2==('m','mi'):
        result = x/meters_per_mile
    elif u1u2==('mi','ft'):
        result = x*5280
    elif u1u2==('ft','mi'):
        result = x/5280
    elif u1u2==('ft','m'):
        result = convert( convert(x,'ft','mi'), 'mi','m')
    elif u1u2==('m','ft'):
        result = convert( convert(x,'m','mi'), 'mi','ft')
        
    # speed
    elif u1u2==('mi/h','m/s'):
        result = convert( convert(x,'mi','m'), 's','h')
    elif u1u2==('m/s','mi/h'):
        result = convert( convert(x,'m','mi'), 'h','s')

    # power 
    elif u1u2==('W','Wh/y'):
        result = convert(x,'y','h')
    elif u1u2==('Wh/y','W'):
        result = convert(x,'h','y')

    # temperature
    elif u1u2==('K','degC'):
        result = x - 273.15/fac 
    elif u1u2==('degC','K'):
        result = x + 273.15/fac 
    elif u1u2==('degC','degF'):
        result = x*9/5 + 32/fac
    elif u1u2==('degF','degC'):
        result = (5/9)*(x-32/fac) 
    elif u1u2==('degF','K'):
        result = convert( convert(x,'degF','degC'), 'degC','K' )
    elif u1u2==('K','degF'):
        result = convert( convert(x,'K','degC'), 'degC','degF' )

    # energy
    elif u1u2==('Wh','BTU'):
        result = x*BTU_per_Wh
    elif u1u2==('BTU','Wh'):
        result = x/BTU_per_Wh
    elif u1u2==('Wh','J'):
        result = convert(x,'h','s')
    elif u1u2==('J','Wh'):
        result = convert(x,'s','h')
    elif u1u2==('eV','J'): # [J] = [V C]
        result = x*eC 
    elif u1u2==('J','eV'):
        result = x/eC 

    # natural units
    elif u1u2==('m','eVinv'):
        result = x/hceVm
    elif u1u2==('eVinv','m'):
        result = x*hceVm
    elif u1u2==('eV','minv'):
        result = x/hceVm
    elif u1u2==('minv','eV'):
        result = x*hceVm
    elif u1u2==('eV','K'):
        result = convert(x,'eV','J')/kBJdivK
    elif u1u2==('K','eV'):
        result = convert(x*kBJdivK,'J','eV')

    else:
        logger.TBRaise('No rule for conversion of ['+u1+'] to ['+u2+']') 

    return fac*result


# This block is to support some legacy code.
hcMeVfm = convert( convert(hceVm,"eV","MeV"), "m","fm" )
hcGeVfm = convert(hcMeVfm,"MeV","GeV")
def fm_to_MeVinv(x) -> float:
    return x/hcMeVfm
def fm_to_GeVinv(x) -> float:
    return x/hcGeVfm
def MeV_to_fminv(x) -> float:
    return x/hcMeVfm
def GeV_to_fminv(x) -> float:
    return x/hcGeVfm
def MeVinv_to_fm(x) -> float:
    return hcMeVfm*x
def GeVinv_to_fm(x) -> float:
    return hcGeVfm*x
def fminv_to_MeV(x) -> float:
    return x*hcMeVfm


class physicalConstant():

    name = None

    def __init__(self,name,scale,units):
        """
        Wrap a dictionary of scale values in physical units.

        Args:
            name (str)
            scale (dict): A dictionary of lists indexed by year. 
            units (str): Base units. See convert method for what units are allowed. 
        """
        checkType(str,name=name)
        checkType(dict,scale=scale)
        if units is not None:
            checkType(str,units=units)
        for key in scale:
            checkType(tuple,scale=scale[key])
        self.name=name
        self.scale=scale
        self.scaleUnits=units
    
    def __repr__(self) -> str:
        if self.name is None:
            return 'physicalConstant'
        return self.name 

    def getValue(self,year,units,returnErr,normalize=1.):
        """
        Retrieve the value of this scale.

        Args:
            year (int): Year this was measured. 
            units (str): Get it in these units. Use None for unitless quantities. 
            returnErr (bool): Return both mean and uncertainty as tuple.
            normalize (real): Return scale/normalize. Defaults to 1.
        """
        checkType('int',year=year)
        if units is not None:
            checkType(str,units=units)
        checkType(bool,returnErr=returnErr)
        checkType("scalar",normalize=normalize)
        if not year in self.scale:
            logger.TBRaise(f"Invalid year specification {year}. Allowed years:",list(self.scale.keys()))
        if self.scaleUnits is None:
            if units is not None:
                logger.TBRaise("Tried to convert a unitless quantity to units",units)
            val = self.scale[year][0]
            err = self.scale[year][1]
        else:
            val = convert(self.scale[year][0],self.scaleUnits,units)/normalize
            err = convert(self.scale[year][1],self.scaleUnits,units)/normalize
        if returnErr:
            return val, err 
        else:
            return val 


# ------------------------------------------------------------------------------------------------------ PARTICLE MASSES 


def M_u_phys(year=2024,units="MeV",returnErr=False):
    scale = {
        2024: (2.16, 0.07), # PDG 2024. DOI: 10.1103/PhysRevD.110.030001
    }
    return physicalConstant("M_u",scale,"MeV").getValue(year,units,returnErr) 


def M_d_phys(year=2024,units="MeV",returnErr=False):
    scale = {
        2024: (4.7, 0.07), # PDG 2024. DOI: 10.1103/PhysRevD.110.030001
    }
    return physicalConstant("M_d",scale,"MeV").getValue(year,units,returnErr) 


def M_s_phys(year=2024,units="MeV",returnErr=False):
    scale = {
        2024: (93.5, 0.8), # PDG 2024. DOI: 10.1103/PhysRevD.110.030001
    }
    return physicalConstant("M_s",scale,"MeV").getValue(year,units,returnErr) 


def M_c_phys(year=2024,units="GeV",returnErr=False):
    scale = {
        2024: (1.273,0.0046), # PDG 2024. DOI: 10.1103/PhysRevD.110.030001
    }
    return physicalConstant("M_c",scale,"MeV").getValue(year,units,returnErr) 


def M_b_phys(year=2024,units="GeV",returnErr=False):
    scale = {
        2024: (4.183,0.007), # PDG 2024. DOI: 10.1103/PhysRevD.110.030001
    }
    return physicalConstant("M_b",scale,"MeV").getValue(year,units,returnErr) 


def M_t_phys(year=2024,units="GeV",returnErr=False):
    scale = {
        2024: (172.57,0.29), # PDG 2024. DOI: 10.1103/PhysRevD.110.030001. direct measurement
    }
    return physicalConstant("M_t",scale,"MeV").getValue(year,units,returnErr) 


def M_mu_phys(year=2022,units="MeV",returnErr=False):
    scale = {
        2022: (105.6583755, 0.0000023), # PDG 2022. DOI: https://doi.org/10.1093/ptep/ptac097.
        2020: (105.6583745, 0.0000024)  # PDG 2020. DOI: https://doi.org/10.1093/ptep/ptaa104.
    }
    return physicalConstant("M_mu",scale,"MeV").getValue(year,units,returnErr) 


def M_pi0_phys(year=2022,units="MeV",returnErr=False):
    scale = {
        2022: (134.9768, 0.0005), # PDG 2022. DOI: https://doi.org/10.1093/ptep/ptac097.
        2014: (134.9766, 0.0006)  # PDG 2014. DOI: 10.1088/1674-1137/38/9/090001
    }
    return physicalConstant("M_pi0",scale,"MeV").getValue(year,units,returnErr) 


def M_pipm_phys(year=2022,units="MeV",returnErr=False):
    scale = {
        2022: (139.57039, 0.00018), # PDG 2022. DOI: https://doi.org/10.1093/ptep/ptac097.
        2014: (139.57018, 0.00035)  # PDG 2014. DOI: 10.1088/1674-1137/38/9/090001
    }
    return physicalConstant("M_pi+/-",scale,"MeV").getValue(year,units,returnErr) 


def M_K0_phys(year=2022,units="MeV",returnErr=False):
    scale = {
        2022: (497.611, 0.013), # PDG 2022. DOI: https://doi.org/10.1093/ptep/ptac097.
    }
    return physicalConstant("M_K0",scale,"MeV").getValue(year,units,returnErr) 


def M_Kpm_phys(year=2022,units="MeV",returnErr=False):
    scale = {
        2022: (493.677, 0.013), # PDG 2022. DOI: https://doi.org/10.1093/ptep/ptac097.
    }
    return physicalConstant("M_K+/-",scale,"MeV").getValue(year,units,returnErr) 


def M_rho_phys(year=2022,units="MeV",returnErr=False):
    scale = {
        2022: (775.26, 0.23), # PDG 2022. DOI: https://doi.org/10.1093/ptep/ptac097.
        2014: (775.26, 0.25)  # PDG 2014. DOI: 10.1088/1674-1137/38/9/090001
    }
    return physicalConstant("M_rho",scale,"MeV").getValue(year,units,returnErr) 


def M_proton_phys(year=2024,units="MeV",returnErr=False):
    scale = {
        2024: (938.27208816, 0.00000029), # PDG 2024. DOI: 10.1103/PhysRevD.110.030001
    }
    return physicalConstant("M_proton",scale,"MeV").getValue(year,units,returnErr) 


def M_neutron_phys(year=2024,units="MeV",returnErr=False):
    scale = {
        2024: (939.56542052, 0.00000054), # PDG 2024. DOI: 10.1103/PhysRevD.110.030001
    }
    return physicalConstant("M_neutron",scale,"MeV").getValue(year,units,returnErr) 



# ------------------------------------------------------------------------------------------------------ DECAY CONSTANTS



def fk_phys(year=2019,units="MeV",returnErr=False):
    """ 
    Physical value of Kaon decay constant, f_K+/-. Scaled by sqrt(2.). 
    """
    scale = {
        2019: (155.7 , 0.7),                     # FLAG 2019. DOI: 10.1140/epjc/s10052-019-7354-7. Section 4.6, Nf=2+1.
        2018: (155.72, quadrature(_fkerrs2018)), # PDG 2018. DOI: 10.1103/PhysRevD.98.030001. Section 84.5.1.
        2012: (156.1 , quadrature(_fkerrs2012)), # PDG 2012. DOI: 10.1103/PhysRevD.86.010001. Page 949 under meson listings.
    }
    return physicalConstant("f_K+/-",scale,"MeV").getValue(year,units,returnErr,normalize=np.sqrt(2.)) 


def fpi_phys(year=2018,units="MeV",returnErr=False):
    """
    Physical value of the pion decay constant, f_pi+/-. 
    """
    scale = {
        2018: (130.50, quadrature(_fpierrs2018)), # PDG 2018. DOI: 10.1103/PhysRevD.98.030001. Section 84.5.1.
    }
    return physicalConstant("f_pi+/-",scale,"MeV").getValue(year,units,returnErr) 


def frho_phys(year=2017,units="GeV",returnErr=False):
    """ 
    Physical value of the rho decay constant. 
    """
    scale = {
        2017: (0.21, 0.01), # HPQCD 2017. DOI: https://doi.org/10.1103/PhysRevD.93.014503. Figure 6.
    }
    return physicalConstant("f_rho",scale,"GeV").getValue(year,units,returnErr) 


# ---------------------------------------------------------------------------------------------------- LATTICE CONSTANTS 


def w0_phys(year=2013,units="fm",returnErr=False):
    """ 
    Gradient flow scale w0.
    """    
    scale = {
        2013: (0.1715, 0.0009), # w0 taken from HPQCD. DOI: 10.1103/PhysRevD.88.074504. Eq (18).
    }
    return physicalConstant("w0",scale,"fm").getValue(year,units,returnErr) 


def r1_phys(year=2010,units="fm",returnErr=False):
    """ 
    Physical value of Sommer scale r1. 
    """    
    scale = {
        2010: (0.3106, quadrature(_r1fmerrs2010)), # r1 taken from MILC 2010. arXiv:1012.0868. 
    }
    return physicalConstant("r1",scale,"fm").getValue(year,units,returnErr) 


_r1phys2010, _r1phys2010e = r1_phys(year=2010,units="fm",returnErr=True)
def r0_phys(year=2014,units="fm",returnErr=False):
    """ 
    Physical value of Sommer scale r0. 
    """    
    scale = {
        2014: (_r1phys2010*1.5092, _r1phys2010e*1.5092), # Use r0/r1 from hotQCD 2014. DOI: https://doi.org/10.1103/PhysRevD.90.094503.
    }
    return physicalConstant("r0",scale,"fm").getValue(year,units,returnErr) 


# ------------------------------------------------------------------------------------------------------ OTHER CONSTANTS 


def alpha_e(year=2018,returnErr=False):
    """ 
    Fine structure constant. 
    """
    scale = {
        2018: (7.2973525693e-3, 0.0000000011e-3), # NIST 2018 CODATA recommended value.
    }
    return physicalConstant("alpha_e",scale,None).getValue(year,None,returnErr) 


def lambda_MSbar_phys(year=2021,units="MeV",returnErr=False):
    """ 
    Physical value of MS-bar lambda parameter. 
    """
    scale = {
        2021: ( 339, 12 ), # Taken from FLAG 2021. arXiv: 2111.09849
    }
    return physicalConstant("lambda_MSbar",scale,"MeV").getValue(year,units,returnErr) 


def Rproton_phys(year=2018,units="fm",returnErr=False):
    """ 
    Physical value of proton charge radius. 
    """
    scale = {
        2018: ( 0.8414, 0.0019 ), # Taken from FLAG 2021. arXiv: 2111.09849
    }
    return physicalConstant("Rproton",scale,"fm").getValue(year,units,returnErr) 



# 
# lattice.py                                                               
# 
# D. Clarke
#
# A basic lattice class for use with statistical mechanical models. 
# 


import numpy as np
import itertools
import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkType


class Lattice():

    """
    A basic lattice class to be used as a framework for statistical mechanics models. This
    is very inefficient at the moment, though easy to read.
    """

    def __init__(self, geometry, example):

        checkType("array",geometry=geometry)

        if np.asarray(example).shape:
            self.nodeShape = np.asarray(example).shape
            self.grid = np.zeros((*geometry, *self.nodeShape))
        else:
            self.nodeShape = 1 
            self.grid = np.zeros(geometry)

        self.Nd       = len(geometry)
        self.geometry = np.array(geometry)
        self.vol      = np.prod(self.geometry)
        self.ranges   = [np.array(range(num)) for num in self.geometry]
        self.bulk     = list(itertools.product(*self.ranges))

        logger.details(f'Initialized {self.Nd}d lattice.')
        logger.details('        dims = ', geometry)
        logger.details('  grid shape = ', self.grid.shape)


    def march(self,coord,mu,step) -> np.ndarray:
        """ 
        Move forward from coordinate coord in direction mu a distance step.

        Args:
            coord (np.array)
            mu (int)
            step (int)

        Returns:
            np.array: New coordinate. 
        """
        newcoord = np.array(coord)
        np.add.at(newcoord, mu, step)
        return newcoord % self.geometry


    def setElement(self,coord,RHS):
        self.grid[tuple(coord)]=RHS


    def getElement(self,coord):
        return self.grid[tuple(coord)]


    # define iterateOverDomain('bulk','random',etc)
    def iterateOverBulk(self,func) -> np.ndarray:
        """ 
        Carry out function func on every site of the lattice. 

        Args:
            func (func): Action to be carried out on every site. Must take coord as
                         its only argument.
        """
        container=[]
        for coord in self.bulk:
            container.append(func(coord))
        return np.array(container)


    def bulkReduce(self,func) -> float:
        return np.sum(self.iterateOverBulk(func))/self.vol


    def iterateOverRandom(self,func) -> np.ndarray:
        """ 
        Carry out function func on every site of the lattice. 

        Args:
            func (func): Action to be carried out on every site. Must take coord as
                         its only argument.
        """
        permutation = np.random.choice(range(self.vol),size=self.vol,replace=False)
        container=[]
        for i in permutation:
            func(self.bulk[i])
        return np.array(container)




# 
# lattice_params.py
# 
# D. Clarke
# 
# Class to handle input parameters of lattice configs. This is in particular for use with the HotQCD collaboration.
#
import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.physics.constants import MeVinv_to_fm, fm_to_MeVinv, fk_phys, r0_phys, r1_phys
from latqcdtools.physics.referenceScales import a_div_r1, a_times_fk, r0_div_a, CY_param, CY_phys
from latqcdtools.base.check import checkDomain


def _massStringToFloat(string):
    if string is None:
        return None
    else:
        return float('0.' + string)


class latticeParams:
    """
    A class to handle and check the input parameters of a lattice run.
    """

    #           mass1  mass2  mass3
    # Nf=1+1+1     mu     md     ms
    # Nf=2+1       ml     ms
    # Nf=X         ml   mpre
    # Nf=2+1+1     ml     ms     mc
    def __init__(self, Nsigma, Ntau, coupling, mass1=None, mass2=None, mass3=None, scaleType='fk', paramYear=None,
                 Nf='21', scaleYear=None, mu=0):
        """ 
        Based on some input, determine all parameters relevant to the ensemble.

        Parameters
        ----------
            Nsigma : int
            Ntau : int
            coupling : str, float
                beta = 6/g^2, where g is the bare coupling.
            mass1 : str
                Light-type mass.
            mass2 : str
                Strange-type mass. For degenerate Nf, can also be an auxiliary mass.
            mass3 : str
                Charm-type mass.
            scaleType : str
                Reference scale.
            paramYear : int
                Year during which the parameters of the rational function or polynomial function used for
                continuum limit extrapolations were determined.
            Nf : str
            scaleYear : int
                Year during which the value of the scale in physical units was determined.
            mu : float
                Baryon chemical potential.
        """
        checkDomain(scaleType,list(CY_phys.keys()))
        self.scale = scaleType
        self.fK    = fk_phys(CY_phys['fk'],"MeV")
        self.r1    = r1_phys(CY_phys['r1'],"fm")
        self.r0    = r0_phys(CY_phys['r0'],"fm")
        if paramYear is None:
            self.year = CY_param[self.scale] 
        else:
            self.year = paramYear
        if scaleYear is not None:
            if self.scale == 'fk':
                self.fK = fk_phys(scaleYear)
            elif self.scale == 'r1':
                self.r1 = r1_phys(scaleYear)
            elif self.scale == 'r0':
                self.r0 = r0_phys(scaleYear)
        if isinstance(coupling, str):
            self.beta  = int(coupling)/10**(len(coupling)-1)
            self.cbeta = coupling
        else:
            self.beta  = coupling
            self.cbeta = str(coupling).replace('.','')
        self.Nc   = 3
        self.mu   = mu
        self.Ns   = Nsigma
        self.Nt   = Ntau
        self.Nf   = Nf
        self.vol4 = self.Ns**3 * self.Nt
        self.cm1  = mass1
        self.cm2  = mass2
        self.cm3  = mass3
        self.cml  = None 
        self.cmu  = None 
        self.cmd  = None 
        self.cms  = None 
        self.ml   = None 
        self.mu   = None 
        self.md   = None 
        self.ms   = None 
        self.cm   = None
        self.cpre = None
        self.m    = None
        self.pre  = None
        self.msml = None
        self.msmu = None
        self.mdmu = None
        if Nf == '21':
            if mass3 is not None:
                logger.TBRaise('Nf=2+1 expects only 2 mass parameters.')
            self.cml  = mass1
            self.cms  = mass2
            self.ml   = _massStringToFloat(mass1)
            self.ms   = _massStringToFloat(mass2)
        elif Nf == '111':
            self.cmu  = mass1 
            self.cmd  = mass2 
            self.cms  = mass3
            self.mu   = _massStringToFloat(mass1)
            self.md   = _massStringToFloat(mass2)
            self.ms   = _massStringToFloat(mass3)
        elif Nf == '211':
            self.cml  = mass1
            self.cms  = mass2
            self.cmc  = mass3
            self.ml   = _massStringToFloat(mass1)
            self.ms   = _massStringToFloat(mass2)
            self.mc   = _massStringToFloat(mass3)
        elif Nf=='3' or Nf=='5':
            if mass3 is not None:
                logger.TBRaise('Degenerate Nf expects only 2 mass parameters.')
            self.cm   = mass1
            self.cpre = mass2
            self.m    = _massStringToFloat(mass1)
            self.pre  = _massStringToFloat(mass2)
        else:
            logger.TBRaise("Unsupported Nf",Nf)
        if (self.ml is not None) and (self.ms is not None):
            self.msml=int(round(self.ms/self.ml))
        if (self.mu is not None) and (self.md is not None):
            self.mdmu=int(round(self.md/self.mu))
        if (self.mu is not None) and (self.ms is not None):
            self.msmu=int(round(self.ms/self.mu))
        if (self.beta<1.) or (10.<self.beta):
            logger.TBRaise("Invalid beta =",self.beta)


    def __repr__(self) -> str:
        return "latticeParams"


    # a in [fm]
    def geta(self):
        if self.scale=='fk':
            return MeVinv_to_fm( a_times_fk(self.beta,self.year)/self.fK )
        elif self.scale=='r1':
            return a_div_r1(self.beta,self.year)*self.r1
        elif self.scale=='r0':
            return self.r0/r0_div_a(self.beta,self.year)


    def getT(self) -> float:
        """ 
        T in MeV. 
        """
        return 1/fm_to_MeVinv( (self.geta()*self.Nt) )


    def getLs(self) -> float:
        """ 
        L in space-like direction in [1/MeV]. 
        """
        if self.Ns is not None:
            return fm_to_MeVinv(self.Ns*self.geta())
        else:
            logger.TBRaise('Must specify Ns get get Ls.')


    # A nicely formatted summary of the lattice parameters.
    def paramSummary(self):
        logger.info()
        logger.info("Lattice parameter summary: ")
        if self.scale == 'fk':
            logger.info("    fK = ",round(self.fK*np.sqrt(2),2),"/sqrt(2) [MeV] ")
        elif self.scale == 'r1':
            logger.info("    r1 = ",self.r1,"[fm] ")
        elif self.scale == 'r0':
            logger.info("    r0 = ",round(self.r0,4),"[fm] ")
        if self.Ns is not None:
            logger.info("    Ns = ",self.Ns)
        logger.info("    Nt = ",self.Nt)
        if self.ml is not None:
            logger.info("    ml = ",self.ml)
        if self.mu is not None:
            logger.info("    mu = ",self.mu)
        if self.md is not None:
            logger.info("    md = ",self.md)
        if self.ms is not None:
            logger.info("    ms = ",self.ms)
        if self.m is not None:
            logger.info("     m = ",self.m)
        if self.pre is not None:
            logger.info("   pre = ",self.pre)
        if self.msml is not None: 
            logger.info(" ms/ml = ",self.msml)
        if self.msmu is not None: 
            logger.info(" ms/mu = ",self.msmu)
        if self.mdmu is not None: 
            logger.info(" md/mu = ",self.mdmu)
        logger.info("    T  = ",round(self.getT(),2), "[MeV]")
        logger.info("    a  = ",round(self.geta(),4), "[fm]")
        if self.Ns is not None:
            logger.info("    Ls = ",round(self.getLs(),4), "1/[MeV]")
        logger.info("  beta = ",self.beta)
        logger.info()



# 
# spline.py                                                               
# 
# D. Clarke
# 
# Generally speaking, one should use scipy's methods for splines, like interp1d, UnivariateSpline, etc. However
# it is a bit inconvenient to use when one wants control over the knots and endpoints. That is what this module is for.
# 

import numpy as np
from scipy.interpolate import CubicSpline, splrep, splev
import latqcdtools.base.logger as logger
from latqcdtools.statistics.statistics import AICc
from latqcdtools.base.check import checkType


def _even_knots(xdata, nknots):
    """ 
    Return a list of nknots evenly spaced knots. 
    """
    if len(xdata)<nknots:
        logger.TBRaise('number of data < number of knots')
    flat_xdata = np.sort(np.asarray(xdata))
    # to ensure no knot sits at a data position
    flat_xdata = np.unique(flat_xdata)
    jump_step = (len(flat_xdata) - 1) / (nknots + 1)
    knots = []
    for i in range(1, nknots + 1):
        x_lower = flat_xdata[int(np.floor(i * jump_step))]
        x_upper = flat_xdata[int(np.ceil(i * jump_step))]
        knots.append((x_lower + x_upper) / 2)
    return knots


def _random_knots(xdata, nknots, randomization_factor=1, SEED=None):
    """ 
    Return a list of nknots randomly spaced knots. 
    """
    rng = np.random.default_rng(SEED)
    flat_xdata = np.sort(np.asarray(xdata))
    sample_xdata = rng.choice(flat_xdata,int(nknots+1+(1-randomization_factor)*(len(flat_xdata)-nknots)),
                              replace=False)
    # Retry if too many data points are removed by np.unique
    if len(np.unique(sample_xdata)) < nknots + 1:
        return _random_knots(xdata, nknots, randomization_factor)
    return _even_knots(sample_xdata, nknots)


class TBSpline:

    """
    A class that prepares a splrep and wraps it with splev.
    """

    def __init__(self,xdata,ydata,edata=None,knots=None,order=3,naturalLike=False):

        self.xspline = np.copy(xdata)
        self.yspline = np.copy(ydata)

        # A "natural spline" is technically a solve that enforces zero curvature at the
        # endpoints. This class on the other hand wraps smoothing splines, usually being
        # applied to data with errors. A strategy to impose zero curvature at the endpoints
        # is to create a fake datum before each endpoint such that the datum is colinear
        # with both the endpoint and the next-innermost point. When there are weights,
        # we weight these last three points more than the rest of the data to try to force
        # the spline to pass through them.
        if naturalLike:
            dxL = xdata[1 ]-xdata[0 ]
            dxR = xdata[-1]-xdata[-2]
            dyL = ydata[1 ]-ydata[0 ]
            dyR = ydata[-1]-ydata[-2]
            self.xspline = np.r_[xdata[0]-dxL, xdata, xdata[-1]+dxR]
            self.yspline = np.r_[ydata[0]-dyL, ydata, ydata[-1]+dyR]

        if edata is None:
            self.weights = None
            smooth = None
        else:
            self.weights = 1/edata
            smooth = len(self.weights)
            if naturalLike:
                weightL = self.weights[0]
                weightR = self.weights[-1] 
                self.weights = np.r_[weightL, self.weights, weightR] 

        self.tck = splrep(self.xspline, self.yspline, t=knots, k=order, w=self.weights, s=smooth, task=-1)

    def __repr__(self) -> str:
        return "TBSpline"

    def __call__(self,x,der=0):
        return splev(x,self.tck,der=der)

    def get_knots(self):
        return self.tck[0]

    def get_coeffs(self):
        return self.tck[1]

    def get_order(self):
        return self.tck[2]

    def get_x(self):
        return self.xspline

    def get_y(self):
        return self.yspline

    def get_weights(self):
        return self.weights


def getSpline(xdata, ydata, num_knots=None, edata=None, order=3, rand=False, fixedKnots=None, 
              getAICc=False, natural=False):
    """ 
    This is a wrapper that calls SciPy spline interpolation methods, depending on your needs. Generally
    this uses scipy.interpolate.splrep, which uses B-splines. If natural=True and edata=None, it will
    use scipy.interpolate.CubicSpline to solve. If natural=True and edata are provided, it will do a
    smoothing spline that attempts to force no curvature at the endpoints, based on the penultimate points. 

    Args:
        xdata (array-like)
        ydata (array-like)
        num_knots (int):
            The number of knots.
        edata (array-like, optional): 
            Error data. Defaults to None.
        order (int, optional):
            Order of the spline. Defaults to 3.
        rand (bool, optional): 
            Use randomly placed knots? Defaults to False.
        fixedKnots (array-like, optional):
            List of user-specified knots. Defaults to None.
        getAICc (bool, optional): 
            Return corrected Aikake information criterion? Defaults to False.
        natural (bool, optional): 
            Try a natural (no change in slope at the endpoints) cubic spline. Defaults to False. 

    Returns:
        callable spline object
        AICc (optionally)
    """

    if len(xdata) != len(ydata):
        logger.TBRaise('len(xdata), len(ydata) =',len(xdata),len(ydata))
    if natural and (order != 3):
        logger.TBRaise("Natural splines have order=3 by definition.")

    if natural and (edata is None): 
        if num_knots is not None:
            logger.TBRaise('Natural spline without error is a solve that chooses knots automatically.')
        spline = CubicSpline(x=xdata,y=ydata,bc_type='natural')

    else:
        checkType('int',num_knots=num_knots)
        nknots = num_knots
        if fixedKnots is not None:
            if type(fixedKnots) is not list:
                logger.TBRaise("knots must be specified as a list.")
            nknots -= len(fixedKnots)
            if nknots < 0:
                logger.TBRaise("len(fixedKnots)",len(fixedKnots),"exceeds num_knots",num_knots)
        if nknots>0:
            if rand:
                knots = _random_knots(xdata,nknots)
            else:
                knots = _even_knots(xdata,nknots)
        else:
            knots=[]
        if fixedKnots is not None:
            for knot in fixedKnots:
                knots.append(knot)
        knots = sorted(knots)
        if knots[0]<xdata[0]:
            logger.TBRaise("You can't put a knot to the left of the x-data. knots, xdata[0] = ",knots,xdata[0])
        if knots[-1]>xdata[-1]:
            logger.TBRaise("You can't put a knot to the right of the x-data. knots, xdata[-1] = ",knots,xdata[-1])
        spline = TBSpline(xdata, ydata, edata=edata, knots=knots, order=order, naturalLike=natural)

    if getAICc:
        cov = np.diag(1/spline.weights**2)
        return spline, AICc(spline.xspline, spline.yspline, cov, spline)
    else:
        return spline


def getSplineErr(xdata, xspline, ydata, ydatae, num_knots=None, order=3, rand=False, fixedKnots=None, natural=False):
    """ 
    Use getSpline to smooth mean and error bars. Create a spline-smooth band from that. 
    """
    spline_lower = getSpline(xdata, ydata - ydatae, num_knots=num_knots, order=order, rand=rand, fixedKnots=fixedKnots, natural=natural)(xspline)
    spline_upper = getSpline(xdata, ydata + ydatae, num_knots=num_knots, order=order, rand=rand, fixedKnots=fixedKnots, natural=natural)(xspline)
    spline_center = getSpline(xdata, ydata, num_knots=num_knots, order=order, rand=rand, fixedKnots=fixedKnots, natural=natural)(xspline)
    spline_err = ((spline_upper - spline_center) + (spline_center - spline_lower)) / 2
    return spline_center, spline_err




# 
# denseObs.py                                                               
# 
# D. Clarke
# 
# Methods to turn output from C. Schmidt's dense code into observables.
# 

import numpy as np
import latqcdtools.base.logger as logger
from latqcdtools.base.readWrite import writeTable
from latqcdtools.base.check import checkType, checkDomain

logger.warn('THIS IS NOT WORKING RIGHT NOW')

_allowed_observables = ["confID",
                        "Nl", "NB", "NQ", "NS", "NI",
                        "dN/dmul", "dN/dmus", "dN/dmuB", "dN/dmuQ", "dN/dmuS", "dN/dmuI",
                        "Nl^2", "NB^2", "NQ^2", "NS^2",
                        "chi2l", "chi2s", "chi11ll", "chi11ls", "chi2B", "chi2Q"]


class observablesOfInterest(list):

    """ 
    A class to specify the dense observables you want to look at. It contains some consistency checks, like making
    sure that you don't add an observable that is not yet computable. It also has some attributes that help streamline
    using methods like np.genfromtxt. 
    """

    def __init__(self,iterable=None):
        if iterable is None:
            iterable = _allowed_observables
        super().__init__(iterable)
        checkType(list,iterable=iterable)
        for item in iterable:
            checkDomain(item,_allowed_observables)
        if not iterable[0] == "confID":
            logger.TBRaise('observablesOfInterest[0] must be confID.')
        self.dtypes=('U12',)
        for item in range(1,len(iterable)):
            self.dtypes += (float,float,)


    def __repr__(self) -> str:
        return "observablesOfInterest"


    def getCol(self,part,obs) -> int:
        """ 
        op_to_obs will give back complex numbers, which are then output to a table. Given an observable obs with
        real or imaginary part part, getCol returns the column index. 
        """
        checkDomain(obs,_allowed_observables)
        if obs=="confID":
            return 0
        if part == "Re":
            return 2*self.index(obs)-1
        elif part == "Im":
            return 2*self.index(obs)
        else:
            logger.TBRaise('part must be Re or Im. Got', part)


def mean_square(vec):
    """ Unbiased calculation of < vec**2 >. """
    N = len(vec)
    #                         diagonal terms
    return ( np.sum(vec)**2 - np.sum(vec**2) )/( N*(N-1))


def op_to_obs(opTable,lp,obs=None,filename='denseObservables.d'):

    """ 
    Take the operators from loadDens and combine them into physically meaningful observables. Some terminology:
        l--light
        s--strange
        B--baryon number
        Q--electric charge
        I--isospin
        S--strangeness

    Parameters
    ----------
    opTable : dict
        A table indexed by confID. Its values are numpy arrays of operators that have been measured.
    lp : latticeParams
        Parameters for the ensemle the configuration belongs to.
    obs : observablesOfInterest, optional
        A list of the observables you want to compute.
    filename : str, optional
        Name for output table.
    """

    logger.warn('This may be wrong. Do not use for now.')

    # Initialize
    if obs is None:
        obs = observablesOfInterest()
    vol4     = lp.vol4
    outTable = {}
    for observable in _allowed_observables:
        outTable[observable] = []

    # Construct the output table
    for cID in opTable:

        if len(cID) != len(cID.strip()):
            logger.TBRaise('confIDs must not have whitespace! This throws off the column indexing.')

        trMdMl=np.array(opTable[cID][0]) # tr M^-1 d M 
        trMdMs=np.array(opTable[cID][1])

        if len(trMdMl) != len(trMdMs): 
            logger.warn("len(trMdMl) != len(trMdMs) cID = "+cID+"... skipping")
            continue

        trMdMl2=np.array(opTable[cID][2])  # tr ( M^-1 d M )^2
        trMdMs2=np.array(opTable[cID][3])

        if len(trMdMl2) != len(trMdMs2):
            logger.warn("len(trMdMl2) != len(trMdMs2) cID = "+cID+"... skipping")
            continue

        trMd2Ml=np.array(opTable[cID][4]) # tr ( M^-1 dd M )^2
        trMd2Ms=np.array(opTable[cID][5])

        if len(trMd2Ml) != len(trMd2Ms): 
            logger.warn("len(trMd2Ml) != len(trMd2Ms), cID = "+cID+"... skipping")
            continue

        if len(trMdMl)==0 or len(trMdMl2)==0 or len(trMd2Ml)==0:
            logger.warn("Found zero random vectors for an observable, cID = "+cID+"... skipping")
            continue

        # I follow the QCD Thermodynamics section of my researchNotes. In the dense code, each trace comes
        # with a 1/vol4. So whenever we have stuff like obs**2, since each factor obs has a trace, we need
        # to multiply by vol4 to a correct normalization. Number densities should be pure imaginary
        # configuration by configuration at mu=0 and for pure imaginary mu, so we extract these pure
        # imaginary parts to reduce the noise. When this quantity is squared, it introduces a (-) sign.
                                                                                            # these should be pure real
        chi2l   = - vol4*( mean_square(trMdMl.imag) )/16 + vol4*np.mean(trMdMl.imag)**2/16 - np.mean(trMdMl2)/4 + np.mean(trMd2Ml)/4
        chi2s   = - vol4*( mean_square(trMdMs.imag) )/16 - np.mean(trMdMs2)/4 + np.mean(trMd2Ms)/4
        chi11ll = - vol4*( mean_square(trMdMl.imag) )/16 + 0*1j 
        chi11ls = - vol4*( np.mean(trMdMl.imag*trMdMs.imag) - np.mean(trMdMl.imag)*np.mean(trMdMs.imag) )/16 + 0*1j
#        chi11ls = - vol4*( np.mean(trMdMl.imag)*np.mean(trMdMs.imag) )/16 + 0*1j

        # TODO: There seems to be a possibility for reuse of this number above 
        nl2  = - mean_square(trMdMl.imag)*vol4/16 + 0j
        ns2  = - mean_square(trMdMs.imag)*vol4/16 + 0j
        nlns = np.mean(trMdMl.imag)*np.mean(trMdMs.imag)*vol4/16 +0j

        chi2Q = (1/9)*( 5*chi2l + chi2s - 4*chi11ll - 2*chi11ls )
        chi2B = (1/9)*( 2*chi2l + chi2s + 2*chi11ll + 4*chi11ls )

        dnl  = -( np.mean(trMdMl2) - np.mean(trMd2Ml) )/4
        dns  = -( np.mean(trMdMs2) - np.mean(trMd2Ms) )/4

        dnS  =           dns
        dnQ  = ( 5*dnl + dns )/9
        dnI  =     dnl + dns
        dn   = ( 2*dnl + dns )/9

        # TODO: possiblity for reuse above?
        nl   =  np.mean( trMdMl )/4
        ns   =  np.mean( trMdMs )/4
        n    =  (2*nl + ns)/3
        nS   = -        ns
        nQ   =  (  nl - ns)/3
        nI   = complex(0)

        nS2  =            ns2
        nQ2  =  (   nl2 + ns2 - 2*nlns )/9
        n2   =  ( 4*nl2 + ns2 + 4*nlns )/9
        nl2  =      nl2

        outTable["confID"].append(cID)
        outTable["Nl"].append(nl)
        outTable["NB"].append(n)
        outTable["NQ"].append(nQ)
        outTable["NS"].append(nS)
        outTable["NI"].append(nI)
        outTable["dN/dmul"].append(dnl)
        outTable["dN/dmus"].append(dns)
        outTable["dN/dmuB"].append(dn )
        outTable["dN/dmuQ"].append(dnQ)
        outTable["dN/dmuS"].append(dnS)
        outTable["dN/dmuI"].append(dnI)
        outTable["Nl^2"].append(nl2)
        outTable["NB^2"].append(n2)
        outTable["NQ^2"].append(nQ2)
        outTable["NS^2"].append(nS2)
        outTable["chi2l"].append(chi2l)
        outTable["chi2s"].append(chi2s)
        outTable["chi11ll"].append(chi11ll)
        outTable["chi11ls"].append(chi11ls)
        outTable["chi2B"].append(chi2B)
        outTable["chi2Q"].append(chi2Q)


    # Prepare the table for file output
    outData   = ()
    outHeader = []
    for observable in obs:
        outData += (outTable[observable],)
        outHeader.append(observable)

    writeTable(filename,*outData,header=outHeader)



#
# num_deriv.py
#
# H. Sandmeyer
#
# Some methods for calculating numerical derivatives. All use the central difference strategy, which has O(h^2)
# corrections, where h is the step size.
#

import numpy as np


def _best_h(x):
    """ 
    This routine attempts to automatically determine the best possible step size h when calculating numerical
    derivatives. One does not like the step size to be too large, since one pays a ~h^2 penalty. On the other hand,
    if the step size is too small, the function may not change within machine precision eps, and hence the derivative
    will be incorrectly estimated.

    Let M_0, M_3 > 0 s.t. |f| < M_0 and |f^(3)| < M_3. Let f_computed = f + f*eps, where eps is the machine precision.
    For central differences,
        ( f(x+h) - f(x-h) )/2h - f' = h^3 f^(3)(x)/3.
    Using the triangle inequality, one can construct an upper bound of the LHS (when calculated on a computer). The h
    which minimizes this upper bound is
        h = (3 eps M_0/M_3)^(1/3).
    Sadly this requires knowledge of M_3, which we presumably don't have since we resorted to numerical derivatives in
    the first place. A compromise that has some memory of this optimization problem is
        h ~ x eps^(1/3).

    We choose eps = 1.1e-16, which is roughly the difference between 1.0 and the next-smallest representable float less
    than 1.0 in 64-bit. The difference between 1.0 and the next-largest float is slightly bigger. 
    """
    eps   = 1.1e-16
    small = pow(eps,1/3)
    h     = small*(abs(x) + small) # Want something meaningful also when x = 0.
    return h


def diff_deriv(x, func, args = (), h = None):
    """ 
    Numerical derivative using central difference. 
    """
    if h is None:
        h = _best_h(x)
    up   = x + h
    down = x - h
    return (func(up, *args) - func(down, *args)) / (2*h)


def diff_grad(params, func, args = (), h = None, floatT=np.float64) -> np.ndarray:
    """ 
    Gradient using difference quotient. 
    """
    ret  = [0.0]*len(params)
    up   = np.array(params, dtype = floatT)
    down = np.array(params, dtype = floatT)
    for i in range(len(params)):
        if h is None:
            h = _best_h(params[i])
        up[i] += h
        down[i] -= h
        ret[i] = (func(up, *args) - func(down, *args)) / (2*h)
        up[i] = params[i]
        down[i] = params[i]
    return np.array(ret,dtype=floatT)


def diff_hess(params, func, args = (), h = None, floatT=np.float64) -> np.ndarray:
    """ 
    Hessian using difference quotient. 
    """

    # This has to be a list, as we might put in arrays, if params is higher dimensional
    ret = [ [0.0]*len(params) for _ in range(len(params)) ]
    # convert to float
    up       = np.array(params, dtype = floatT)
    down     = np.array(params, dtype = floatT)
    upup     = np.array(params, dtype = floatT)
    updown   = np.array(params, dtype = floatT)
    downdown = np.array(params, dtype = floatT)
    downup   = np.array(params, dtype = floatT)

    for i in range(len(params)):
        if h is None:
            hi = _best_h(params[i])
        else:
            hi = h
        up[i] += hi
        upup[i] = up[i]
        updown[i] = up[i]
        down[i] -= hi
        downdown[i] = down[i]
        downup[i] = down[i]
        ret[i][i]=(func(up,*args)+func(down,*args)-2*func(params,*args))/(4*(hi/2.0)**2)

        for j in range(i):
            if h is None:
                hj = _best_h(params[j])
            else:
                hj=h

            upup[j]+=hj
            updown[j]-=hj
            downdown[j]-=hj
            downup[j]+=hj

            ret[i][j]=(func(upup,*args) + func(downdown,*args)-func(updown,*args) - func(downup,*args)) / (4*hi*hj)
            ret[j][i]=ret[i][j]

            upup[j]=params[j]
            updown[j]=params[j]
            downdown[j]=params[j]
            downup[j]=params[j]

        up[i] = params[i]
        upup[i] = up[i]
        updown[i] = up[i]
        down[i] = params[i]
        downdown[i] = down[i]
        downup[i] = down[i]

    return np.array(ret,dtype=floatT)


def diff_fit_grad(x, params, func, args = (), h = None):
    """ 
    When fitting we're trying to optimize params, and hence we want to think of func as a function of
    its parameters rather than x. 
    """ 
    def f(p):
        return func(x, p, *args)
    return diff_grad(params, f, h = h)


def diff_fit_hess(x, params, func, args = (), h = None):
    """ 
    When fitting we're trying to optimize params, and hence we want to think of func as a function of
    its parameters rather than x. 
    """ 
    def f(p):
        return func(x, p, *args)
    return diff_hess(params, f, h = h)


def diff_jac(params, func, args = (), h = None) -> np.ndarray:
    return diff_grad(params, func, args, h).transpose()



# 
# HotQCD.py
# 
# D. Clarke
# 
# Some parameter combinations and naming conventions special to projects of the HotQCD collaboration.
#

import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkType
from latqcdtools.math.math import rel_check
from latqcdtools.interfaces.interfaces import HotQCD_MILC_Params


class HotQCDParams(HotQCD_MILC_Params):
    """
    A class to handle and check the input parameters of a lattice run, especially for HotQCD.
    """

    def __repr__(self) -> str:
        return "HotQCDParams"

    def getcGradFlowPureGauge(self) -> str:
        return 's'+str(self.Ns).zfill(3)+'t'+str(self.Nt).zfill(2)+'_b0'+self.cbeta+'00'


def makeConfTag(conf,stream) -> str:
    """ 
    This takes a configuration number conf and stream label stream to make a tag labelling a configuration.
    Implementing this as a function makes sure everyone using the Toolbox as the same convention and, more importantly,
    ensures that the tags have no whitespace in them, which otherwise can throw off the column counting of methods in
    the denseObs module. 
    """
    return f'{stream}:{conf}'


def loadDens(densFile,confID,lp,inTable=None) -> dict:
    """ 
    Allows reading of output from C. Schmidt's Dense code. The Dense code produces measurements of various operators
    relevant for calculating conserved charge fluctuations. We store as a dictionary indexed by confID, which lets us
    conveniently combine data when there are multiple dense files per configuration. Here we update the table
    inTable for the new read-in, which yields outTable. Only supports Nf=2+1 configurations for the time being.

    Parameters
    ----------
    densFile : str
        Name of the Dense code file to be read.
    confID : str
        A unique identifier for the configuration on which dense calculation was performed. Every person seems to have a
        different scheme for labelling configurations, so this needs to be a string to be as flexible as possible.
    lp : latticeParams
        Parameters for the ensemle the configuration belongs to.
    inTable : dict
        A table indexed by confID. Its values are a list of operators that have been measured.

    Returns
    -------
    outTable : dict
        A table indexed by confID. Its values are a list of operators that have been measured.
    """

    if len(confID) != len(confID.strip()):
        logger.TBRaise('confID must not contain whitespace.')

    if inTable is None:
        outTable  = {}
    else:
        checkType(dict,inTable=inTable)
        outTable = inTable

    try:
        infile = open(densFile, 'r')
    except Exception as e:
        logger.warn("Unable to open file",densFile)
        raise e

    logger.info("Loading file",densFile)
    Nc = lp.Nc

    # In the following light and strange quarks are indexed by l and s, respectively.
    if not confID in outTable:
        nlVec      = []  # tr M^-1 d M 
        nsVec      = []
        nl2Vec     = []  # tr ( M^-1 d M )^2
        ns2Vec     = []
        MddMlVec   = []  # tr M^-1 dd M
        MddMsVec   = []
        trMinvlVec = []  # tr M^-1
        trMinvsVec = []
    else:
        nlVec      = outTable[confID][0]
        nsVec      = outTable[confID][1]
        nl2Vec     = outTable[confID][2]
        ns2Vec     = outTable[confID][3]
        MddMlVec   = outTable[confID][4]
        MddMsVec   = outTable[confID][5]
        trMinvlVec = outTable[confID][6]
        trMinvsVec = outTable[confID][7]

    # Read in data from the dense file.
    lineno = 0
    for line in infile:

        lineno += 1
        col = line.split()

        # Parse the densFile.
        try:
            OPID = int(col[0])
            mass = float(col[2])
            ReOP = float(col[3])
            ImOP = float(col[4])
        except Exception as e:
            infile.close()
            logger.warn("Read error on line", lineno, "of file", densFile)
            raise e

        # If you have a trace squared, you must normalize by volume. Each trace must be normalized by Nc.
        if OPID == 1:  # tr M^-1
            if rel_check(mass, lp.ml):
                trMinvlVec.append(Nc * complex(ReOP, ImOP))
            elif rel_check(mass, lp.ms):
                trMinvsVec.append(Nc * complex(ReOP, ImOP))
            else:
                logger.TBRaise("Unexpected mass on line", lineno, "of file", densFile, ". ms, ml, m =", lp.ms, lp.ml, mass)

        elif OPID == 2:  # D1 = tr M^-1 d M
            if rel_check(mass, lp.ml):
                nlVec.append(Nc * complex(ReOP, ImOP))
            elif rel_check(mass, lp.ms):
                nsVec.append(Nc * complex(ReOP, ImOP))
            else:
                logger.TBRaise("Unexpected mass on line", lineno, "of file", densFile, ". ms, ml, m =", lp.ms, lp.ml, mass)

        elif OPID == 3:  # tr M^-1 dd M
            if rel_check(mass, lp.ml):
                MddMlVec.append(Nc * complex(ReOP, ImOP))
            elif rel_check(mass, lp.ms):
                MddMsVec.append(Nc * complex(ReOP, ImOP))
            else:
                logger.TBRaise("Unexpected mass on line", lineno, "of file", densFile, ". ms, ml, m =", lp.ms, lp.ml, mass)

        elif OPID == 12:  # tr (M^-1 d M)**2
            if rel_check(mass, lp.ml):
                nl2Vec.append(Nc * complex(ReOP, ImOP))
            elif rel_check(mass, lp.ms):
                ns2Vec.append(Nc * complex(ReOP, ImOP))
            else:
                logger.TBRaise("Unexpected mass on line", lineno, "of file", densFile, ". ms, ml, m =", lp.ms, lp.ml, mass)

        else:
            continue

    outTable[confID] = [nlVec, nsVec, nl2Vec, ns2Vec, MddMlVec, MddMsVec, trMinvlVec, trMinvsVec]

    infile.close()

    return outTable



#
# plotting.py
#
# H. Sandmeyer, D. Clarke
#
# Collection of convenience tools for plotting using matplotlib.
#


import itertools, os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
import latqcdtools.base.logger as logger
from latqcdtools.base.check import checkEqualLengths, checkType
from latqcdtools.base.utilities import isHigherDimensional, toNumpy , envector, createFilePath
from latqcdtools.base.readWrite import readTable

try:
    if 'DISPLAY' not in os.environ:
        # Check if there's no display (headless environment)
        matplotlib.use('Agg')
    else:
        # This is to avoid possible problems with colors and themes. As far as I can tell,
        # everything works with PyQt5. 
        matplotlib.use('Qt5Agg') 
except ModuleNotFoundError:
    pass


ZOD        = 10     # Global, Orders different layers 
INITIALIZE = True   # Global, flag to ensure we only initialize once
LEGEND     = False  # Global, flag to apply legend attributes when we have one
TWINX      = None   # Global, legend info is accumulated in this axis 
FOREGROUND = 99999  # zorder = FOREGROUND to put something in plot on top of everything else 
BACKGROUND = 0      # zorder = BACKGROUND to put something in plot behind everything else 


colors = ['#d32d11', '#0081bf', '#e5af11', '#7c966d', '#7570b3', '#ff934f', '#666666', '#D186B3']


markers_1 = ['o', 'v', 'D', 's', 'p', '^', 'h', 'd', 'x', '+', '*']
markers   = itertools.cycle(markers_1)


# Handles are what graphics you're going to use to represent a particular legend element. Usually
# these are taken from some feature of the plotted data. The labels are the names they will have.
# We implement these as dictionaries to allow the user to have different legend objects in
# for instance multiple subplots.
legend_handles = { plt : [] }
legend_labels  = { plt : [] }


default_params = {

    'ax': plt,    # Axis object that is used for the plots. Default is matplotlib.pyplot.
    'ZOD': None,  # Controls where in foreground/background data/lines/bands appear.

    # Basic options affecting most plots.
    'xlabel': None,
    'alpha_xlabel': 0,           # Transparency for x-label
    'ylabel': None,
    'alpha_ylabel': 0,           # Transparency for y-label
    'xmin': None,
    'xmax': None,
    'ymin': None,
    'ymax': None,
    'title': None,
    'label': None,               # What are the data called? (Will appear in legend.)
    'color': None,               # Color for your data. (By default each new set automatically gets different color.)
    'facecolor' : 'color',       # Sometimes one differentiates between colors of faces and edges. (By default same as color.)
    'font_size': 12,             # Controls font size for everything. 
    'font_weight': 'normal',     # Default style of font ('normal', 'bold', 'heavy', 'light')
    'alpha': 0.5,                # General transparency for data.
    'ticksintoplot': True,       # Put ticks into plotting area.
    'surroundWithTicks': True,   # Put ticks also on top and right.
    'labelsintoplot': True,      # Put xlabel and ylabel into plotting area.
    'xlabelpos': None,           # If labelsintplot=True, shift the position (x,y) of the x-label, expressed as percent.
    'ylabelpos': None,
    'alpha_dots': None,          # Transperancy for different dots
    'alpha_lines': 1,            # Transperancy for different lines
    'alpha_fill_edge': 0,        # Transperancy for edges of error bands
    'hatch': None,               # Fill pattern
    'linewidth': 1,              # Linewidth of line plots
    'capsize': 1.5,              # Length of caps af error bars
    'elinewidth': 1,             # Linewidth of the error bars of caps af error bars
    'grid' : False,              # Do you want to put a grid in the background?

    # Data markers
    'marker': "iter",            # Symbol used for plotting data. (Set to 'None' if you don't want any.)
    'markersize': 8,             # Size of the symbols.
    'markerfill': False,         # If False, markers are hollow

    # Options for the legend. 
    # legendpos:
    #   2     9     1           'upper left'    'upper center'  'upper right'
    #   6     10    7           'center left'   'center'        'center right'
    #   3     8     4           'lower left'    'lower center'  'lower right'
    'legendpos': 'best',
    'bbox_to_anchor': None,      # Manual position of the legend. If this is not None, then legendpos
                                 #   controls the point on the legend that bbox_to_anchor controls.
                                 #   for example legendpos='lower left' along with bbox_to_anchor=(0,0)
                                 #   puts the lower left corner of the legend at the bottom left corner
                                 #   of the plot. The same legendpos with bbox_to_anchor=(1,1) puts the
                                 #   lower left corner of the legend in the upper right corner of the plot.
                                 #   This is not my fault--this was a matplotlib design choice.
    'legend_ncol': 1,            # Number of columns in the legend.
    'legend_col_spacing': None,  # Spacing between columns in the legend.
    'legendframeon' : True,      # Thin border around the legend.
    'legend_title': None,        # Title of the legend.
    'handletextpad': 0.2,        # Spacing between symbol and text in legend.
    'alpha_legend': 1,           # Transperancy for the legend.

    # Adjust special aspects of the plot's axes
    'xtick_freq': None,
    'ytick_freq': None,
    'xtick_every_n': 2,    # If xtick_freq or ytick_freq is not None, label every nth tick.
    'ytick_every_n': 2,
    'xscale': 1.0,         # Scale data in xdata by this factor.
    'yscale': 1.0,
    'xlogscale': False,    # Should we use a log scale for the x-axis?
    'ylogscale': False,
    'xlogbase': 10,
    'ylogbase': 10,
    'orientation' : 'vertical',
}


# Used for later checks that the user did not pass a wrong parameter by mistake.
allowed_params = set(default_params.keys())
allowed_params = allowed_params | {'linestyle','ha','va'}


# ---------------------------------------------------------------------------------------------- SOME EXTERNAL FUNCTIONS


def latexify(bold=False):
    """ 
    Allows use of LaTeX symbols in plots. The physics package is included, allowing use of
    convenient functions like ev. 
    """
    try:
        logger.info("Using LaTeX to make pretty plots.")
        if bold:
            plt.rcParams['text.latex.preamble'] = r"\usepackage{lmodern}\usepackage{amssymb}\usepackage{physics}\boldmath"
        else:
            plt.rcParams['text.latex.preamble'] = r"\usepackage{lmodern}\usepackage{amssymb}\usepackage{physics}"
        plt.rcParams['text.usetex'] = True
        plt.rcParams['font.family'] = 'cmr10'
        plt.rcParams['axes.formatter.use_mathtext'] = True
    except Exception as e:
        logger.warn("You seem to be missing some LaTeX packages, so latexify won't work. Encountered exception:")
        logger.warn(e)


def resetLEGEND():
    global LEGEND, TWINX
    LEGEND = False
    TWINX = None


def clearPlot():
    """ 
    Clears plot object, legend handles, and zorder. Useful if you want to do multiple plots in the same script. 
    """
    logger.debug('Reset plot defaults.')
    global INITIALIZE, ZOD, LEGEND, legend_labels, legend_handles
    ZOD = 10
    INITIALIZE = True
    resetLEGEND() 
    legend_handles = { plt : [] }
    legend_labels  = { plt : [] }
    plt.clf()
    plt.close('all')


def getColorGradient(NUM_COLORS,map='viridis') -> list:
    """ 
    Generate perceptually uniform set of colors. Useful when you need more than 8 colors.

    Args:
        NUM_COLORS (int): number of colors you need 
        map (str, optional): use custom colormap. Defaults to 'viridis'.

    Returns:
        list: colors 
    """
    checkType('int',NUM_COLORS=NUM_COLORS)
    checkType(str,map=map)
    cm = plt.get_cmap(map)
    gradColors=[]
    for i in range(NUM_COLORS):
        color = cm(1.*i/NUM_COLORS)
        gradColors.append(color)
    return gradColors


def set_xrange(xmin=None, xmax=None, ax=plt):
    _set_xmin(ax,xmin)
    _set_xmax(ax,xmax)


def set_yrange(ymin=None, ymax=None, ax=plt):
    _set_ymin(ax,ymin)
    _set_ymax(ax,ymax)


def set_default_param(**kwargs):
    """ 
    Lets the user adjust the default parameter settings. 
    """
    for key, val in kwargs.items():
        default_params[key] = val


def getSubplots(x,y):
    """ 
    Get fig and axs objects when you want x figures across and y figures vertically.
    I wrapped this because matplotlib's convention for x and y is the opposite as their
    convention for figsize, which is so incredibly confusing. 
    
    Args:
        x (int): number of panels in x-direction 
        y (int): number of panels in y-direction 

    Returns:
        fig, axs: fig object, list (if 1-d) of ax objects or tuple (if 2-d)
    """
    checkType('int',x=x)
    checkType('int',y=y)
    fig, axs = plt.subplots(y,x,figsize=(4*x,4*y))
    return fig, axs


def getTwinAxes():
    """
    Get two axes objects ax1 and ax2 that share the same x-axis. Useful for graphing two functions
    on the same plot with different scales.
    """
    global TWINX
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()
    TWINX = ax2
    return ax1, ax2


# ---------------------------------------------------------------------------------------------- SOME INTERNAL FUNCTIONS


def _initializePlt(params):
    """ 
    Set up inital plot parameters, like its size. I tried to introduce a global variable INITIALIZE that checks
    so that this only gets called once per plot. 
    """
    checkType(dict,params=params)
    fill_param_dict(params)
    global INITIALIZE
    if INITIALIZE:
        logger.debug("Plot initializer called!")
        INITIALIZE = False
        plt.rcParams['savefig.bbox']      = 'standard'
        plt.rcParams['axes.titlesize']    = params['font_size']
        plt.rcParams['font.size']         = params['font_size']
        plt.rcParams['ytick.labelsize']   = params['font_size']
        plt.rcParams['xtick.labelsize']   = params['font_size']
        plt.rcParams['axes.labelsize']    = params['font_size']
        plt.rcParams['legend.fontsize']   = params['font_size']
        plt.rcParams['font.weight']       = params['font_weight']
        set_xrange(params['xmin'],params['xmax'])
        set_yrange(params['ymin'],params['ymax'])


def _update_labels(ax,label):
    checkType(str,label=label)
    global legend_labels, TWINX
    if ax in legend_labels:
        legend_labels[ax].append(label)
    else:
        legend_labels[ax] = [label]
    if (TWINX is not None) and (ax != TWINX): 
        if TWINX in legend_labels:
            legend_labels[TWINX].append(label)
        else:
            legend_labels[TWINX] = [label]


def _update_handles(ax,handle):
    global legend_handles, TWINX
    if ax in legend_handles:
        legend_handles[ax].append(handle)
    else:
        legend_handles[ax] = [handle]
    if (TWINX is not None) and (ax != TWINX): 
        if TWINX in legend_handles:
            legend_handles[TWINX].append(handle)
        else:
            legend_handles[TWINX] = [handle]


def _prepare_legend(ax,params):
    """
    Calls set_params, which adds labels and prepares the plot for display.
    """
    global TWINX
    if TWINX is None:
        set_params(**params)
    else:
        if ax==TWINX:
            set_params(**params)


def _getAxObject(params):
    if params['ax']==plt:
        return plt.gca()
    else:
        return params['ax']


def _getZOD(params):
    ZOD = params['ZOD']
    if ZOD is None:
        ZOD = globals()['ZOD']
    return ZOD


def _set_xmin(ax,x_min=None):
    if x_min is not None:
        try:
            x1, x2 = ax.get_xlim()
            ax.set_xlim([x_min,x2])
        except AttributeError:
            # The issue here is that the get_xlim() won't know what x2
            # is without having seen the data first.
            logger.TBRaise('Must set x/y min/max after plotting data.',exception=AttributeError)


def _set_xmax(ax,x_max=None):
    if x_max is not None:
        try:
            x1, x2 = ax.get_xlim()
            ax.set_xlim([x1,x_max])
        except AttributeError:
            logger.TBRaise('Must set x/y min/max after plotting data.',exception=AttributeError)


def _set_ymin(ax,y_min=None):
    if y_min is not None:
        try:
            y1, y2 = ax.get_ylim()
            ax.set_ylim([y_min,y2])
        except AttributeError:
            logger.TBRaise('Must set x/y min/max after plotting data.',exception=AttributeError)


def _set_ymax(ax,y_max=None):
    if y_max is not None:
        try:
            y1, y2 = ax.get_ylim()
            ax.set_ylim([y1,y_max])
        except AttributeError:
            logger.TBRaise('Must set x/y min/max after plotting data.',exception=AttributeError)


def _rescale(scale,data):
    """ 
    Rescale the data, taking into account there may be no data.

    Args:
        scale (float)
        data (array-like or None)
    """
    if data is not None:
        return np.copy(data*scale)


def fill_param_dict(params):
    """ 
    Collection of default parameters for plotting routines. If a key does not exist in params, it is defined with
    a default value.
    
    Args:
        params (dict): All parameters that are already set by the user
    """
    global LEGEND

    if not LEGEND:
        # When filling params, check if we show the legend. This is triggered by one of these keys
        # being different from the default value. 
        for key in ('legend_title', 'legendpos', 'legend_ncol', 'legendpos_col_spacing', 'label', 
                    'alpha_legend', 'bbox_to_anchor', 'legendframeon'):
            if key in params:
                if params[key] != default_params[key]:
                    logger.debug('Found legend trigger',key)
                    LEGEND = True

    for key, val in default_params.items():
        params.setdefault(key,val)


def _add_optional(params) -> dict:
    """ 
    Optional parameters not included in _fill_param_dict.

    Args:
        **params: Additional parameters that can be set.

    Returns:
        dict: also optional parameters 
    """
    reference = {}
    fill_param_dict(reference)
    ret = {}
    for key in params:
        if not key in reference:
            ret[key] = params[key]
    return ret


def set_params(**params):
    """ 
    Set additional parameters to the plot. For example set a title or label.

    Args:
        **params: Additional parameters that can be set.
    """
    global LEGEND

    fill_param_dict(params)
    ax = _getAxObject(params)

    # Check for some contradictory settings.
    if (params['xlabelpos'] is not None) and (params['xlabel'] is None):
        logger.warn('Set xlabelpos with no xlabel.')
    if (params['ylabelpos'] is not None) and (params['ylabel'] is None):
        logger.warn('Set ylabelpos with no ylabel.')
    if params['xlogscale'] and params['xtick_freq'] is not None:
        logger.warn("xtick_freq assumes no log scale.")
    if params['ylogscale'] and params['ytick_freq'] is not None:
        logger.warn("ytick_freq assumes no log scale.")

    if params['xlabel'] is not None:
        checkType(str,xlabel=params['xlabel'])
        if params['labelsintoplot'] or params['xlabelpos'] is not None:
            if params['xlabelpos'] is None:
                params['xlabelpos'] = (0.95,0.027)
            ax.annotate(params['xlabel'], xy=params['xlabelpos'], xycoords='axes fraction', color='black',
                        fontsize=params['font_size'], fontweight=params['font_weight'], ha='right', va='bottom',
                        bbox=dict(linewidth=0, facecolor='white', edgecolor=None, alpha=params['alpha_xlabel']), 
                        zorder=FOREGROUND)
        else:
            ax.set_xlabel(params['xlabel'])
            ax.xaxis.get_label().set_fontsize(params['font_size'])

    if params['ylabel'] is not None:
        checkType(str,ylabel=params['ylabel'])
        if params['labelsintoplot'] or params['ylabelpos'] is not None:
            if params['ylabelpos'] is None:
                params['ylabelpos'] = (0.025,0.972)
            ax.annotate(params['ylabel'], xy=params['ylabelpos'], xycoords='axes fraction', color='black',
                        fontsize=params['font_size'], ha='left', va='top', fontweight=params['font_weight'],
                        bbox=dict(linewidth=0, facecolor='white', edgecolor=None, alpha=params['alpha_ylabel']), 
                        zorder=FOREGROUND)
        else:
            ax.set_ylabel(params['ylabel'])
            ax.yaxis.get_label().set_fontsize(params['font_size'])

    ax.tick_params(axis='both', which='major', labelsize=params['font_size']) 

    if params['title'] is not None:
        checkType(str,title=params['title'])
        plt.title(params['title'])

    if params['xlogscale']:
        ax.set_xscale('log',base=params['xlogbase'])

    if params['ylogscale']:
        ax.set_yscale('log',base=params['ylogbase'])

    if params['ticksintoplot']:
        ax.tick_params(which='both',direction='in')

    if params['surroundWithTicks']:
        ax.tick_params(which='minor',left=True,bottom=True,top=True,right=True)
        ax.tick_params(which='major',left=True,bottom=True,top=True,right=True)

    if params['grid']:
        ax.grid(True, color='lightgray',linestyle='dotted',zorder=BACKGROUND)

    set_xrange(params['xmin'],params['xmax'],ax)
    set_yrange(params['ymin'],params['ymax'],ax)

    if LEGEND:
        if not ax in legend_handles:
            logger.TBRaise('Legend for axis',ax,'was activated without any label.')
        leg = ax.legend(legend_handles[ax], legend_labels[ax], numpoints=1, bbox_to_anchor = params['bbox_to_anchor'],
                        title=params['legend_title'], loc=params['legendpos'], ncol=params['legend_ncol'],
                        columnspacing=params['legend_col_spacing'],handletextpad = params['handletextpad'],
                        frameon = params['legendframeon'])
        leg.get_frame().set_alpha(params['alpha_legend'])
        leg.set_zorder(FOREGROUND)

    if params['xtick_freq'] is not None:
        start, end = ax.get_xlim()
        if params['xmin'] is not None:
            start = params['xmin']
        if params['xmax'] is not None:
            end = params['xmax']
        ax.xaxis.set_ticks(np.arange(start, end, params['xtick_freq']))
        for n, label in enumerate(ax.xaxis.get_ticklabels()):
            if n % params['xtick_every_n'] != 0:
                label.set_visible(False)

    if params['ytick_freq'] is not None:
        start, end = ax.get_ylim()
        if params['ymin'] is not None:
            start = params['ymin']
        if params['ymax'] is not None:
            end = params['ymax']
        ax.yaxis.set_ticks(np.arange(start, end, params['ytick_freq']))
        for n, label in enumerate(ax.yaxis.get_ticklabels()):
            if n % params['ytick_every_n'] != 0:
                label.set_visible(False)


# ------------------------------------------------------------------------------------------------ MAIN PLOTTING METHODS


def preliminary(x,y,text='PRELIMINARY',**kwargs):
    """ 
    Generate a PRELIMINARY tag on the plot.

    Args:
        x (float): x-position of bottom-left corner (in units of x-axis) 
        y (float): y-position of bottom-left corner (in units of y-axis) 
        text (str, optional): Text indicating result is preliminary. Defaults to 'PRELIMINARY'.
    """
    checkType(str,text=text)
    _initializePlt(kwargs)
    if 'color' in kwargs: 
        color=kwargs['color']
    else:
        color='gray'
    axObj=plt
    if 'ax' in kwargs:
        axObj=kwargs['ax']
    axObj.text(x, y, text, color=color)


def plot_file(filename, xcol=0, ycol=1, yecol=None, xecol=None, func = None, func_args = (), style='dots', **params):
    """ 
    Plot data in file. You can set the style with the style argument. Columns indexed from 0.

    Args:
        filename (str)
        xcol (int, optional): Which column is xdata. Defaults to 1.
        ycol (int, optional): Which column is ydata. Defaults to 2.
        yecol (int, optional): Which column has y error. Defaults to None.
        xecol (int, optional): Which column has x error. Defaults to None.
        func (function, optional): Apply this function to the data. Defaults to None.
        func_args (tuple, optional): Arguments to func. Defaults to ().
        style (str, optional): Choose from dots, lines, fill, and band. Defaults to 'dots'.
        **params: Additional parameters that can be set.
    """
    checkType('int',xcol=xcol)
    checkType('int',ycol=ycol)
    _initializePlt(params)
    data   = readTable(filename,dtype=str)
    xdata  = data[xcol].astype(float)
    ydata  = data[ycol].astype(float)
    yedata = None
    xedata = None

    if yecol is not None:
        yedata = data[yecol].astype(float)
    else:
        if style == "fill":
            logger.TBRaise("Need error column for filled plotting")
    if xecol is not None:
        xedata = data[xecol].astype(float)

    checkEqualLengths(xdata,ydata,xedata,yedata)

    if func is not None:
        if yedata is not None:
            if xedata is not None:
                xdata, ydata, yedata, xedata = func(xdata, ydata, yedata, xedata, *func_args)
            else:
                xdata, ydata, yedata = func(xdata, ydata, yedata, *func_args)
        else:
            xdata, ydata = func(xdata, ydata, *func_args)

    if style == "dots":
        plot_dots(xdata, ydata, yedata=yedata, xedata=xedata, **params)
    elif style == "lines":
        plot_lines(xdata, ydata, yedata=yedata, xedata=xedata, **params)
    elif style == "fill":
        plot_fill(xdata, ydata, yedata=yedata, **params)
    else:
        logger.TBRaise("Unknown style",style)


def plot_vspan(minVal,maxVal,**params):
    """ 
    Plot a vertical band.

    Args:
        minVal (float)
        maxVal (float)
    """
    _initializePlt(params)
    optional = _add_optional(params)
    ax       = _getAxObject(params)
    ZOD      = _getZOD(params)
    handle   = ax.axvspan(xmin=minVal, xmax=maxVal,color=params['color'],alpha=params['alpha'],zorder=ZOD,**optional)
    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,handle)
        _prepare_legend(ax,params)
    globals()['ZOD'] += 1


def plot_hspan(minVal,maxVal,**params):
    """ 
    Plot a horizontal band.

    Args:
        minVal (float)
        maxVal (float)
    """
    _initializePlt(params)
    optional = _add_optional(params)
    ax       = _getAxObject(params)
    ZOD      = _getZOD(params)
    handle   = ax.axhspan(ymin=minVal, ymax=maxVal,color=params['color'],alpha=params['alpha'],zorder=ZOD,**optional)
    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,handle)
        _prepare_legend(ax,params)
    globals()['ZOD'] += 1


def plot_hline(y,minVal=None,maxVal=None,**params):
    """ 
    Plot a horizontal line at y. 
    """
    _initializePlt(params)
    optional = _add_optional(params)
    ax       = _getAxObject(params)
    ZOD      = _getZOD(params)
    x1, x2   = ax.get_xlim()
    if minVal is not None:
        x1=minVal
    if maxVal is not None:
        x2=maxVal
    logger.debug('x1, x2 = ',x1,x2)
    # The matplotlib people are psychopaths. I couldn't get this to work in any other
    # way. Yes, the comma after handle in the else branch is mandatory. 
    if (minVal is None) and (maxVal is None):
        handle = ax.axhline(y=y,color=params['color'],zorder=ZOD,alpha=params['alpha'], **optional) 
    else:
        handle, = ax.plot([x1, x2], [y, y], color=params['color'],zorder=ZOD,alpha=params['alpha'], **optional) 
    globals()['ZOD'] += 1
    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,handle)
        _prepare_legend(ax,params)


def plot_vline(x,minVal=None,maxVal=None,**params):
    """ 
    Plot a vertical line at x. 
    """
    _initializePlt(params)
    optional = _add_optional(params)
    ax       = _getAxObject(params)
    ZOD      = _getZOD(params)
    y1, y2   = ax.get_ylim()
    if minVal is not None:
        y1=minVal
    if maxVal is not None:
        y2=maxVal
    logger.debug('y1, y2 = ',y1,y2)
    if (minVal is None) and (maxVal is None):
        handle = ax.axvline(x=x,color=params['color'],zorder=ZOD,alpha=params['alpha'], **optional) 
    else:
        handle, = ax.plot([x, x], [y1, y2], color=params['color'], zorder=ZOD, alpha=params['alpha'], **optional) 
    globals()['ZOD'] += 1
    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,handle)
        _prepare_legend(ax,params)


def plot_dots(xdata, ydata, yedata = None, xedata = None, **params):
    """ 
    Plot ydata vs xdata as dots. 

    Args:
        xdata (array-like)
        ydata (array-like)
        yedata (array-like, optional): y error. Defaults to None.
        xedata (array-like, optional): x error. Defaults to None.
        **params: Additional parameters that can be set.
    """

    if len(envector(yedata)) == 2:
        checkEqualLengths(xdata,ydata,yedata.T)
    else:
        checkEqualLengths(xdata,ydata,yedata)

    if len(envector(xedata)) == 2:
        checkEqualLengths(xdata,ydata,xedata.T)
    else:
        checkEqualLengths(xdata,ydata,xedata)

    xdata, ydata, xedata, yedata = toNumpy(xdata, ydata, xedata, yedata)
    _initializePlt(params)
    optional = _add_optional(params)

    # If you want lines between your data you should be using plot_lines.
    if 'linestyle' in optional:
        logger.warn('Ignoring linestyle',optional['linestyle'])
        del optional['linestyle']

    ax     = _getAxObject(params)
    xedata = _rescale(abs(params['xscale']),xedata)
    yedata = _rescale(abs(params['yscale']),yedata)
    ZOD    = _getZOD(params)

    marker = params['marker']
    if marker == "iter":
        marker = next(markers)
    if params['markerfill']:
        markerfill=params['color']
    else:
        markerfill="None"

    if params['color'] is not None:
        handle = ax.errorbar(xdata*params['xscale'], ydata*params['yscale'], yerr=yedata, xerr=xedata, marker=marker, 
                           linestyle='None', linewidth=params['linewidth'], alpha=params['alpha'], color=params['color'],
                           zorder=ZOD, markersize=params['markersize'], capsize=params['capsize'],
                           elinewidth=params['elinewidth'],
                           markerfacecolor = markerfill, **optional)
    else:
        handle = ax.errorbar(xdata*params['xscale'], ydata*params['yscale'], yerr=yedata, xerr=xedata, marker=marker, 
                           linestyle='None', linewidth=params['linewidth'], alpha=params['alpha'], zorder=ZOD,
                           markersize=params['markersize'], capsize=params['capsize'], elinewidth=params['elinewidth'],
                           markerfacecolor = markerfill, **optional)

    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,handle)
        _prepare_legend(ax,params)

    globals()['ZOD'] += 1


def plot_bar(xdata, ydata, width=None, align='edge', edgecolor='#666677',linewidth=0.2, **params):
    """ 
    Plot ydata vs xdata as bars.

    Args:
        xdata (array-like)
        ydata (array-like)
        width (float, optional): Width of bar. Defaults to xdata[1]-xdata[0].
        align (str, optional): How to align the bars. Defaults to 'edge'.
        edgecolor (str, optional): Color of bar edges. Defaults to '#666677'.
        linewidth (float, optional): Defaults to 0.2.
        **params: Additional parameters that can be set.
    """
    checkEqualLengths(xdata,ydata)
    xdata, ydata = toNumpy(xdata, ydata)
    _initializePlt(params) 
    optional = _add_optional(params)
    ax       = _getAxObject(params)
    ZOD      = _getZOD(params)

    if width is None:
        width = xdata[1] - xdata[0]

    if params['color'] is not None:
        bar = ax.bar(xdata, ydata, color=params['color'], zorder=ZOD, width=width, align=align, edgecolor=edgecolor,
                     linewidth=linewidth, alpha=params['alpha'], **optional)
    else:
        bar = ax.bar(xdata, ydata, zorder=ZOD, width=width, align=align, edgecolor=edgecolor,
                     linewidth=linewidth, alpha=params['alpha'], **optional)

    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,bar)
        _prepare_legend(ax,params)
    globals()['ZOD'] += 1


def plot_hist(data, bins = None, density=False, label=None, weights=None, **params):
    """ 
    Create a histogram of the array data. If you would like to plot multiple data sets in the same histogram,
    simply pass as a list or tuple of arrays of data, like data = [list1, list2, ...].

    Args:
        data (array-like)
        bins (int, optional): Number of bins. Defaults to None, which sets the number of bins automatically.
        **params: Additional parameters that can be set.
    """
    _initializePlt(params) 
    ax = _getAxObject(params)
    if bins is None:
        bins = 'auto'
        if weights is not None:
            logger.TBRaise('Matplotlib does not support automatic bins with weights')
    if isHigherDimensional(data):
        ax.hist(data, bins=bins, density=density, label=label, orientation=params['orientation'], alpha=params['alpha'],
                weights=weights)
    else:
        ax.hist(data, bins=bins, density=density, label=label, color=params['color'], orientation=params['orientation'],
                alpha=params['alpha'], weights=weights)
    if density:
        ax.set_yticklabels([])
    if label is not None:
        ax.legend()


def plot_lines(xdata, ydata, yedata=None, xedata=None, **params):
    """ 
    Plot ydata vs xdata using lines.

    Args:
        xdata (array-like)
        ydata (array-like)
        yedata (array-like, optional): y error. Defaults to None.
        xedata (array-like, optional): x error. Defaults to None.
        **params: Additional parameters that can be set.
    """
    checkEqualLengths(xdata,ydata,yedata,xedata)
    xdata, ydata, xedata, yedata = toNumpy(xdata, ydata, xedata, yedata)
    _initializePlt(params)
    optional = _add_optional(params)
    ax       = _getAxObject(params)
    ZOD      = _getZOD(params)

    xedata = _rescale(params['xscale'],xedata)
    yedata = _rescale(params['yscale'],yedata)

    marker = params['marker']
    if marker == "iter":
        marker = next(markers)
    if params['markerfill']:
        markerfill=params['color']
    else:
        markerfill="None"

    if params['color'] is not None:
        handle = ax.errorbar(xdata*params['xscale'], ydata*params['yscale'], yerr=yedata, xerr=xedata, marker=marker, 
                           linestyle='None', linewidth=params['linewidth'], color=params['color'], zorder=ZOD,
                           markersize=params['markersize'], capsize=params['capsize'], elinewidth=params['elinewidth'],
                           markeredgewidth=params['elinewidth'], alpha=params['alpha_dots'], 
                           markerfacecolor = markerfill)
    else:
        handle = ax.errorbar(xdata*params['xscale'], ydata*params['yscale'], yerr=yedata, xerr=xedata, marker=marker, 
                           linestyle='None', linewidth=params['linewidth'], zorder=ZOD, markersize=params['markersize'],
                           capsize=params['capsize'], elinewidth=params['elinewidth'], 
                           markeredgewidth=params['elinewidth'], alpha=params['alpha_dots'],
                           markerfacecolor = markerfill)

    globals()['ZOD'] += 1

    col = handle[0].get_color()

    line = ax.errorbar(xdata*params['xscale'], ydata*params['yscale'], color = col, linewidth=params['linewidth'], 
                       zorder = ZOD, alpha = params["alpha"], **optional)

    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,(line,handle))
        _prepare_legend(ax,params)


def plot_fill(xdata, ydata, yedata, xedata=None, center=True, **params):
    """ 
    Plot a filled region within ydata +/- yedata. Can set xedata along with yedata=None for vertical bands.

    Args:
        xdata (array-like)
        ydata (array-like)
        yedata (array-like): y error 
        xedata (array-like, optional): x error. Defaults to None.
        center (bool): Do you show the central line? Defaults to True. 
        **params: Additional parameters that can be set.
    """
    if (yedata is None) and (xedata is None):
        logger.TBRaise("Please pass some error bars.")
    if (yedata is not None) and (xedata is not None):
        logger.TBRaise("Please pass either x-error or y-error, not both.")
    xdata, ydata, xedata, yedata = toNumpy(xdata, ydata, xedata, yedata)

    if len(envector(yedata)) == 2:
        checkEqualLengths(xdata,ydata,xedata,yedata.T)
    else:
        checkEqualLengths(xdata,ydata,xedata,yedata)

    _initializePlt(params) 
    optional = _add_optional(params)
    ax       = _getAxObject(params)
    ZOD      = _getZOD(params)

    if center :
        if params['color'] is not None:
            handle = ax.errorbar(xdata*params['xscale'], ydata*params['yscale'], linewidth=params['linewidth'], 
                                 color=params['color'], zorder=ZOD+1, alpha = params['alpha_lines'], **optional)
        else:
            handle = ax.errorbar(xdata*params['xscale'], ydata*params['yscale'], linewidth=params['linewidth'], 
                                 zorder=ZOD+1, alpha = params['alpha_lines'], **optional)
        col = handle[0].get_color()
    else:
        col = params['color']

    if params['facecolor'] == 'color':
        facecol=col
    else:
        facecol=params['facecolor']

    if xedata is None:
        if len(yedata) == 2:
            pl = ax.fill_between(xdata*params['xscale'],
                                 (np.asarray(ydata*params['yscale']) - np.asarray(yedata[0]*params['yscale'])),
                                 (np.asarray(ydata*params['yscale']) + np.asarray(yedata[1]*params['yscale'])), 
                                 facecolor=facecol, alpha=params['alpha'], linewidth=params['linewidth'], 
                                 zorder=ZOD, edgecolor=col, hatch=params['hatch'])
        else:    
            pl = ax.fill_between(xdata*params['xscale'],
                                 (np.asarray(ydata*params['yscale']) - np.asarray(yedata*params['yscale'])),
                                 (np.asarray(ydata*params['yscale']) + np.asarray(yedata*params['yscale'])), 
                                 facecolor=facecol, alpha=params['alpha'], linewidth=params['linewidth'], 
                                 zorder=ZOD, edgecolor=col, hatch=params['hatch'])
    else:
        pl = ax.fill_betweenx(ydata*params['yscale'], 
                              (np.asarray(xdata*params['xscale']) - np.asarray(xedata*params['xscale'])),
                              (np.asarray(xdata*params['xscale']) + np.asarray(xedata*params['xscale'])), 
                              facecolor=facecol, alpha=params['alpha'],linewidth=params['linewidth'], 
                              zorder=ZOD, edgecolor=col, hatch=params['hatch'])
    globals()['ZOD'] += 2

    if params['label'] is not None:
        _update_labels(ax,params['label'])
        _update_handles(ax,(handle,pl))
        _prepare_legend(ax,params)


def plot_matrix(mat,vmin=None,vmax=None):
    """ Plot matrix as a heatmap.

    Args:
        mat (np.ndarray): correlation matrix
        ax (matplotlib ax object): Defaults to plt.
    """
    checkType(np.ndarray,mat=mat)
    heatmapColors = ['blue', 'white', 'red']
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list('custom_cmap', heatmapColors)
    plt.imshow(mat, cmap=cmap, vmin=vmin,vmax=vmax)
    plt.colorbar()


def saveFigure(filename,**kwargs):
    """
    Wrapper for plt.savefig that creates the directory path if it doesn't exist already.
    """
    createFilePath(filename)
    plt.savefig(filename,**kwargs)
